
from collections import defaultdict
from copy import copy
from logging import getLogger
from re import compile, sub

from urwid import Pile, Text, Filler, Edit, Columns, Frame, Divider, Padding, connect_signal

from ch2.diary.database import COMPARE_LINKS
from ch2.diary.model import TYPE, VALUE, TEXT, DP, HI, LO, FLOAT, UNITS, SCORE, LABEL, EDIT, MEASURES, SCHEDULES, TAG, \
    LINK, INTEGER, DB, value, text
from ch2.lib import format_seconds
from ch2.lib.utils import format_watts, format_percent, format_metres
from ch2.stats.display.nearby import NEARBY_LINKS
from ch2.stats.names import S, W, PC, M
from ch2.urwid.tui.decorators import Border, Indent
from ch2.urwid.tui.fixed import Fixed
from ch2.urwid.tui.tabs import Tab
from ch2.urwid.tui.widgets import Float, Rating0, Rating1, ArrowMenu, DividedPile, SquareButton, Integer

log = getLogger(__name__)

HR_ZONES_WIDTH = 32  # characters total

COLUMN_WIDTH = 6     # characters per column
N_COLUMNS = 12       # columns on screen (typically use multiple; 12 allows 3 or 4 'real' columns)


# for text markup
def em(text): return 'em', text
def error(text): return 'error', text
def label(text): return 'label', text
def zone(zone, text): return 'zone-%d' % zone, text
def quintile(quintile, text): return 'quintile-%d' % quintile, text


def build(model, f, layout):
    footer = ['meta-', em('q'), 'uit/e', em('x'), 'it/', em('s'), 'ave']
    footer += [' [shift]meta-']
    for sep, c in enumerate('dwmya'):
        if sep: footer += ['/']
        footer += [em(c)]
    footer += ['ctivity/', em('t'), 'oday']
    active = defaultdict(list)
    branch = layout(model, f, active)
    return active, Border(Frame(Filler(branch, valign='top'),
                                footer=Pile([Divider(), Text(footer, align='center')])))


def apply_before(model, f, active, before, after, leaf):
    key = model[0].get(TAG, None)
    return before[key](key, model, f, active, copy(before), copy(after), copy(leaf))


def layout(model, f, active, before, after, leaf):
    '''
    Takes a model and returns an urwid widget.
    The model is nested lists, forming a tree, with nodes that are dicts.
    This function traverses the tree in a depth-first manner, converting dicts using leaf and assembling
    the visited nodes using after.
    The before map can be used to intercept normal processing.
    Before and after are keyed on the 'tag' entry in the first dict in the list;
    leaf is keyed on the 'type' entry in the dict.
    '''

    if isinstance(model, list):
        if not model:
            raise Exception('Empty list in model')
        if not isinstance(model[0], dict):
            raise Exception(f'Model list with no head element: {model}')
        try:
            key, branch = apply_before(model, f, active, before, after, leaf)
            return after[key](branch)
        except Exception as e:
            log.error(f'Error ({e}) while processing {model}')
            raise
    else:
        if not isinstance(model, dict):
            raise Exception(f'Model entry of type {type(model)} ({model})')
        key = model.get(TYPE, None)
        try:
            return leaf[key](model, f, active)
        except Exception as e:
            log.error(f'Error ({e}) while processing leaf {key} in {model}')
            raise


def fmt_value_units(model):
    if model[UNITS] == S:
        return [format_seconds(model[VALUE])]
    elif model[UNITS] == W:
        return [format_watts(model[VALUE])]
    elif model[UNITS] == PC:
        return [format_percent(model[VALUE])]
    else:
        text = []
        if isinstance(model[VALUE], float):
            if 1 < model[VALUE] < 1000:
                text += ['%.1f' % model[VALUE]]
            else:
                text += ['%g' % model[VALUE]]
        else:
            text += [str(model[VALUE])]
        if model[UNITS]:
            text += [model[UNITS]]
        return text


def fmt_value_measures(model):
    measures = []
    if MEASURES in model and model[MEASURES]:
        for schedule in model[MEASURES][SCHEDULES]:
            percentile, rank = model[MEASURES][SCHEDULES][schedule]
            q = 1 + min(4, percentile / 20)
            measures.append(quintile(q, f'{int(percentile)}%:{rank}/{schedule} '))
    if not measures:
        measures = [label('--')]
    return measures


def create_value(model, f, active):
    return Text([label(model[LABEL] + ': ')] + fmt_value_units(model) + [' '] + fmt_value_measures(model))


def create_value_no_measures(model, f, active):
    return Text([label(model[LABEL] + ': ')] + fmt_value_units(model))


def create_link(model, f, active):
    button = SquareButton(model[VALUE], state=model[DB])
    active[model[TAG]].append(button)
    return f(Padding(Fixed(button, len(model[VALUE]) + 2), width='clip'))


def wire(model, widget):
    def callback(w, v):
        model[DB].value = v
    connect_signal(widget, 'change', callback)
    return widget


def null_leaf(model, f, active):
    return model


def default_leaf(model, f, active):
    raise Exception(f'Unexpected leaf: {model}')


LEAF_DATE = defaultdict(
    lambda: default_leaf,
    {
        TEXT: lambda model, f, active: Text(model[VALUE]),
        EDIT: lambda model, f, active: f(wire(model,
                                              Edit(caption=label(model[LABEL] + ': '), edit_text=model[VALUE] or ''))),
        FLOAT: lambda model, f, active: f(wire(model,
                                               Float(caption=label(model[LABEL] + ': '), state=model[VALUE],
                                                     minimum=model[LO], maximum=model[HI], dp=model[DP],
                                                     units=model[UNITS]))),
        INTEGER: lambda model, f, active: f(wire(model,
                                                 Integer(caption=label(model[LABEL] + ': '), state=model[VALUE],
                                                         minimum=model[LO], maximum=model[HI], units=model[UNITS]))),
        SCORE: lambda model, f, active:  f(wire(model,
                                                Rating0(caption=label(model[LABEL] + ': '), state=model[VALUE]))),
        VALUE: create_value,
        LINK: create_link
    })


LEAF_SCHEDULE = defaultdict(
    lambda: default_leaf,
    {
        TEXT: lambda model, f, active: Text(model[VALUE]),
        VALUE: null_leaf,
        LINK: create_link
    })


def menu(key, model, f, active, before, after, leaf):
    menu = ArrowMenu(label(model[0][VALUE] + ': '),
                     {link[DB]: link[VALUE] for link in model[1:]})
    active[key].append(menu)
    return key, f(menu)


def side_by_side(*specs):

    def before(key, model, f, active, before, after, leaf):
        branch_columns = []
        for names in specs:
            try:
                columns, reduced_model, found = [], list(model), False
                for name in names:
                    for i, m in enumerate(reduced_model):
                        if isinstance(m, list) and isinstance(m[0], dict) and m[0].get(TAG) == name:
                            columns.append(m)
                            del reduced_model[i]
                            found = True
                            break
                    if not found:
                        raise Exception(f'Missing column {name}')
                columns = [layout(column, f, active, before, after, leaf) for column in columns]
                branch_columns.append(Columns(columns))
                model = reduced_model
            except Exception as e:
                log.warning(e)
        branch = [layout(m, f, active, before, after, leaf) for m in model]
        branch.extend(branch_columns)
        return key, branch

    return before


def value_to_row(value, has_measures=None):
    # has_measures can be true/false to force
    row = [Text(label(value[LABEL])), Text(fmt_value_units(value))]
    if has_measures is True or (MEASURES in value and SCHEDULES in value[MEASURES] and value[MEASURES][SCHEDULES]):
        row += [Text(fmt_value_measures(value))]
    elif has_measures is None:
        row += [Text('')]
    return row


def rows_to_table(rows):
    widths = [max(len(row.text) for row in column) for column in zip(*rows)]
    return Columns([(width+1, Pile(column)) for column, width in zip(zip(*rows), widths)])


def values_table(key, model, f, active, before, after, leaf):
    values = [m for m in model if isinstance(m, dict) and m.get(TYPE, None) == 'value']
    rest = [layout(m, f, active, before, after, leaf) for m in model if m not in values]
    table = rows_to_table([value_to_row(value) for value in values])
    return key, rest + [table]


def climbs_table(key, model, f, active, before, after, leaf):

    def climb(model):
        # assumes title, elevation, distance and time entries, in that order
        return [Text(fmt_value_units(model[1])), Text(fmt_value_units(model[2])), Text(fmt_value_units(model[3])),
                Text(fmt_value_measures(model[1]))]

    elevations = [m for m in model if isinstance(m, list) and m[0].get(TAG, None) == 'climb']
    rest = [layout(m, f, active, before, after, leaf) for m in model if m not in elevations]
    table = rows_to_table([climb(elevation) for elevation in elevations])
    return key, rest[:1] + [table] + rest[1:]  # title, table, total


def table(name, value):

    def before(key, model, f, active, before, after, leaf):
        title = layout(model[0], f, active, before, after, leaf)
        has_measures = any(MEASURES in m for m in model[1:])
        headers = [name, value]
        if has_measures: headers.append('Stats')
        rows = [[Text(label(header)) for header in headers]] + [value_to_row(m, has_measures) for m in model[1:]]
        return key, [title, rows_to_table(rows)]

    return before


def shrimp_table(key, model, f, active, before, after, leaf):

    def reformat(model):
        branch = [Text(label(model[0][VALUE])), Text(str(model[1][VALUE])), Text(em(model[3][VALUE])),
                  Text(str(model[2][VALUE]))]
        for ranges in model[4:]:
            branch.append(Text(label(f'{ranges[1][VALUE]}-{ranges[2][VALUE]}/{ranges[0][TAG]}')))
        return branch

    return key, [Text(model[0][VALUE]), rows_to_table([reformat(m) for m in model[1:]])]


def collapse_title(key, model, f, active, before, after, leaf):
    def is_text(x):
        return isinstance(x, dict) and x[TYPE] == TEXT
    if len(model) == 2 and is_text(model[0]) and len(model[1]) and is_text(model[1][0]):
        model[1][0][VALUE] = f'{model[0][VALUE]} - {model[1][0][VALUE]}'
        model = model[1]
    else:
        del before[key]
    return apply_before(model, f, active, before, after, leaf)


def hr_zone(key, model, f, active, before, after, leaf):
    width = HR_ZONES_WIDTH
    z = model[0][VALUE]
    pc = model[1][VALUE]
    text = ('%d:' + ' ' * (width - 6) + '%3d%%') % (z, int(0.5 + pc))
    column = 100 / width
    left = int((pc + 0.5 * column) // column)
    text_left = text[0:left]
    text_right = text[left:]
    return key, Text([zone(z, text_left), text_right])


def schedule_combine(*patterns):

    patterns = [compile(pattern) for pattern in patterns]

    def before(key, model, f, active, before, after, leaf):

        def rows():
            current_pattern, columns = None, []
            for row in model:
                if not is_row(row):
                    yield row
                else:
                    if not current_pattern:
                        for pattern in patterns:
                            title = row[0][VALUE]
                            match = pattern.match(title)
                            if match:
                                current_pattern = pattern
                                i = title.find(match.group(1))
                                title = sub(r'\s+', ' ', title[:i] + title[(i + len(match.group(1))):])
                                columns = [text(title)]
                    if current_pattern:
                        match = current_pattern.match(row[0][VALUE])
                        if match:
                            columns.append(value(match.group(1), row[1][VALUE], units=row[1][UNITS]))
                        else:
                            yield columns
                            current_pattern, columns = None, []
                    else:
                        yield row
            if columns:
                yield columns

        if len(model) > 1 and any(is_row(b) for b in model[1:]):
            model = list(rows())
        return key, [layout(m, f, active, before, after, leaf) for m in model]

    return before


def drop_measures(key, model, f, active, before, after, leaf):
    leaf[VALUE] = create_value_no_measures
    return default_before(key, model, f, active, before, after, leaf)


def default_before(key, model, f, active, before, after, leaf):
    if not isinstance(model, list):
        raise Exception(f'"before" called with non-list type ({type(model)}, {model})')
    return key, [layout(m, f, active, before, after, leaf) for m in model]


BEFORE_DATE = defaultdict(
    lambda: default_before,
    {'activity': side_by_side(('hr-zones-time', 'climbs'),
                              ('min-time', 'med-time'),
                              ('max-med-heart-rate', 'max-mean-power-estimate')),
     'min-time': table('Dist', 'Time'),
     'med-time': table('Dist', 'Time'),
     'max-med-heart-rate': table('Time', 'HR'),
     'max-mean-power-estimate': table('Time', 'Power'),
     'activity-statistics': values_table,
     'segments': collapse_title,
     'segment': values_table,
     'climbs': climbs_table,
     'shrimp': shrimp_table,
     'nearbys': collapse_title,
     'hr-zone': hr_zone,
     'database': drop_measures,
     NEARBY_LINKS: menu,
     COMPARE_LINKS: menu
     })


BEFORE_SCHEDULE = defaultdict(
    lambda: schedule_combine(r'Min (\d+km) Time', r'Med (\d+km) Time', r'Max Med HR (\d+m)'),
    {})


def widget_size(widget, max_cols=N_COLUMNS):
    if isinstance(widget, Tab): widget = widget.w
    for cls in (Float, Rating0, Rating1):
        if isinstance(widget, cls): return 3
    if isinstance(widget, ArrowMenu): return 6
    if isinstance(widget, Text) and not isinstance(widget, Edit):
        return min(max_cols, 1 + len(widget.text) // COLUMN_WIDTH)
    return max_cols


def pack_widgets(branch, max_cols=N_COLUMNS):
    new_branch = [branch[0]]
    columns, width = [], 0
    for widget in branch[1:]:
        size = widget_size(widget)
        if size + width <= max_cols:
            columns.append((COLUMN_WIDTH * size, widget))
            width += size
        else:
            if columns:
                new_branch.append(Columns(columns))
            columns, width = [(COLUMN_WIDTH * size, widget)], size
    if columns:
        new_branch.append(Columns(columns))
    return default_after(new_branch)


def title_after(branch):
    head, tail = branch[0], branch[1:]
    if tail:
        return Pile([head, Divider(), Indent(DividedPile(tail))])
    else:
        return head


def null_after(branch):
    return branch


def is_row(branch):
    # check for row in schedule format
    return isinstance(branch, list) and len(branch) > 1 and isinstance(branch[1], dict) and branch[1][TYPE] == VALUE


def schedule_tables(branch):

    def fmt_rows(branch, ncols, max_cols=4):
        row = [branch[0]]
        for b in branch[1:]:
            # extra case to handle up/down arrow in fitness
            row.append(create_value_no_measures(b, None, None) if isinstance(b, dict) else b)
            if len(row) == max_cols:
                yield row
                row = [Text('')]
        if len(row) > 1:
            while len(row) < max_cols: row.append(Text(''))
            yield row

    if is_row(branch):
        return branch  # will be processed below
    if isinstance(branch, list) and len(branch) > 1 and any(is_row(b) for b in branch[1:]):
        ncols = max(len(b) for b in branch[1:] if is_row(b))
        all_rows, table = [], []
        for b in branch[1:]:
            if is_row(b):
                for row in fmt_rows(b, ncols): table.append(row)
            else:
                if table:
                    all_rows.append(Columns([Pile(col) for col in zip(*table)]))
                    table = []
                all_rows.append(b)
        if table:
            all_rows.append(Columns([Pile(col) for col in zip(*table)]))
        return Pile([branch[0], Indent(Pile(all_rows))])
    else:
        return default_after(branch)


def default_after(branch):
    head, tail = branch[0], branch[1:]
    if tail:
        return Pile([head, Indent(Pile(tail))])
    else:
        return head


AFTER_DATE = defaultdict(
    lambda: default_after,
    {'status': pack_widgets,
     'nearby': pack_widgets,
     'title': title_after,
     'hr-zone': null_after,
     NEARBY_LINKS: null_after,
     COMPARE_LINKS: null_after})


AFTER_SCHEDULE = defaultdict(
    lambda: schedule_tables,
    {'title': title_after})


def layout_date(model, f, active, before=BEFORE_DATE, after=AFTER_DATE, leaf=LEAF_DATE):
    return layout(model, f, active, before=before, after=after, leaf=leaf)


def layout_schedule(model, f, active, before=BEFORE_SCHEDULE, after=AFTER_SCHEDULE, leaf=LEAF_SCHEDULE):
    return layout(model, f, active, before=before, after=after, leaf=leaf)




from logging import getLogger

from . import MultiProcCalculator, DataFrameCalculatorMixin, SegmentJournalCalculatorMixin
from ..names import SEGMENT_TIME, S, summaries, MIN, MSR, CNT, HEART_RATE, SEGMENT_HEART_RATE, BPM, MAX
from ...data import activity_statistics, present, linear_resample_time
from ...sql import SegmentJournal, StatisticJournalFloat

log = getLogger(__name__)

SJOURNAL = 'sjournal'


class SegmentCalculator(SegmentJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def _delete(self, s):
        super()._delete(s)
        SegmentJournal.clean(s)

    def _read_dataframe(self, s, sjournal):
        return activity_statistics(s, HEART_RATE, activity_journal=sjournal.activity_journal)

    def _calculate_stats(self, s, sjournal, df):
        stats = {SJOURNAL: sjournal,
                 SEGMENT_TIME: (sjournal.finish - sjournal.start).total_seconds()}
        if present(df, HEART_RATE):
            ldf = linear_resample_time(df, start=sjournal.start, finish=sjournal.finish)
            stats[SEGMENT_HEART_RATE] = ldf[HEART_RATE].mean()
        return stats

    def _copy_results(self, s, ajournal, loader, stats):
        sjournal = stats[SJOURNAL]
        loader.add(SEGMENT_TIME, S, summaries(MIN, CNT, MSR), sjournal.segment, sjournal,
                   stats[SEGMENT_TIME], sjournal.start, StatisticJournalFloat)
        if SEGMENT_HEART_RATE in stats:
            loader.add(SEGMENT_HEART_RATE, BPM, summaries(MAX, CNT, MSR), sjournal.segment, sjournal,
                       stats[SEGMENT_HEART_RATE], sjournal.start, StatisticJournalFloat)




from functools import lru_cache
from genericpath import exists
from logging import getLogger

from math import floor
from os.path import join
from zipfile import ZipFile

import numpy as np

from ..lib.log import log_current_exception
from ..sql import Constant

log = getLogger(__name__)


SRTM1_DIR = 'SRTM1.Dir'
SAMPLES = 3601
# from view-source:http://dwtkns.com/srtm30m/
BASE_URL = 'http://e4ftl01.cr.usgs.gov/MEASURES/SRTMGL1.003/2000.02.11/'
EXTN = '.SRTMGL1.hgt.zip'


# lots of credit to https://github.com/aatishnn/srtm-python/blob/master/srtm.py
# (although that has bugs...)


@lru_cache(4)  # 4 means our tests are quick (and should tile a local patch)
def cached_file_reader(dir, flat, flon):
    # https://wiki.openstreetmap.org/wiki/SRTM
    # The official 3-arc-second and 1-arc-second data for versions 2.1 and 3.0 are divided into 1°×1° data tiles.
    # The tiles are distributed as zip files containing HGT files labeled with the coordinate of the southwest cell.
    # For example, the file N20E100.hgt contains data from 20°N to 21°N and from 100°E to 101°E inclusive.
    root = '%s%02d%s%03d' % ('S' if flat < 0 else 'N', abs(flat), 'W' if flon < 0 else 'E', abs(flon))
    hgt_file = root + '.hgt'
    hgt_path = join(dir, hgt_file)
    zip_path = join(dir, root + EXTN)
    if exists(hgt_path):
        log.debug(f'Reading {hgt_path}')
        with open(hgt_path, 'rb') as input:
            data = input.read()
    elif exists(zip_path):
        log.debug(f'Reading {zip_path}')
        with open(zip_path, 'rb') as input:
            zip = ZipFile(input)
            log.debug(f'Found {zip.filelist}')
            data = zip.open(hgt_file).read()
    else:
        # i tried automating download, but couldn't get ouath2 to work
        log.warning(f'Download {BASE_URL + root + EXTN}')
        raise Exception(f'Missing {hgt_file}')
    return np.flip(np.frombuffer(data, np.dtype('>i2'), SAMPLES * SAMPLES).reshape((SAMPLES, SAMPLES)), 0)


class ElevationSupport:

    def __init__(self, dir, reader=cached_file_reader):
        self._dir = dir
        self._reader = reader

    def _lookup(self, lat, lon):
        flat, flon = floor(lat), floor(lon)
        # construct the path in the reader so it's skipped if we hit the cache
        return flat, flon, self._reader(self._dir, flat, flon)


def elevation_from_constant(s, interp, dir_name=SRTM1_DIR):
    try:
        dir = Constant.get(s, dir_name).at(s).value
        if not exists(dir): raise Exception(f'SRTM1 directory {dir} missing')
    except:
        log_current_exception(traceback=False)
        log.warning(f'STRM1 config - define {dir_name} in constants for elevation data')
        dir = None
    return interp(dir)




from math import exp

from ..lib import to_time
from ..sql import StatisticJournal, Constant, ActivityGroup
from ..stats.names import FTHR

# values from british cycling online calculator
# these are upper limits
BC_ZONES = (68, 83, 94, 105, 121)


# these functions were used during development, but are not used in current code
# which does calculations on dataframes for efficiency.
# they have been left for documentation (and possibly useful from jupyter).


def zone(hr, fthr, zones=BC_ZONES):
    zones = [x * fthr / 100.0 for x in zones]
    for i in range(len(zones)):
        if hr < zones[i]:
            if i == len(zones) - 1:
                return i + 1 + (hr - zones[i-1]) / (zones[i-1] - zones[i-2])
            elif i:
                return i + 1 + (hr - zones[i-1]) / (zones[i] - zones[i-1])
            else:
                return 2 + (hr - zones[0]) / (zones[1] - zones[0])


def shrimp(hr, gamma, zero, fthr, zones=BC_ZONES):
    hrz = zone(hr, fthr, zones)
    n = len(zones) - 1
    return n * (max(0, hrz - zero) / (n - zero)) ** gamma


def trimp(hr, rest_hr, max_hr, k):
    x = (hr - rest_hr) / (max_hr - rest_hr)
    return x * 0.64 * exp(k * x)


def edwards(hr, max_hr):
    return max(0, int(10 * (1 - (max_hr - hr) / max_hr)) - 4)


def hr_zones_from_database(s, local_time, activity_group_name):
    activity_group = ActivityGroup.from_name(s, activity_group_name)
    fthr = StatisticJournal.before(s, to_time(local_time), FTHR, Constant, activity_group)
    if fthr:
        return hr_zones(fthr.value)
    else:
        return None


def hr_zones(fthr):
    return [fthr * pc / 100.0 for pc in BC_ZONES]




import rasterio as rio
from matplotlib.pyplot import gca

from ch2.data import *
from ch2.lib.image import write_image, overlay_route, matplot_route, matplot_image
from ch2.msil2a import *
from ch2.msil2a.elevation import create_elevation, add_elevation
from ch2.jupyter.decorator import template


@template
def route_sentinel_3d(user, passwd, local_time, activity_group_name):

    f'''
    # Route : {local_time} ({activity_group_name})

    Generate a 3D landscape with the route marked in red.
    The terrain image is taken from Sentinel satellite data.
    The elevation model is calculated from SRTM data (already used to give activity elevation).
    Currently (see below) this cannot be displayed in the notebook, but is saved to file as GeoTIFF data.
    '''

    '''
    $contents
    '''

    '''
    ## Download Image Data
    '''

    s = session('-v2')

    api, products, bbox, df = query_activity(s, user, passwd, local_time, activity_group_name)
    download_paths = cached_download(s, api, products)

    '''
    ## Create Image
    '''
    image_paths = [create_rgb(download) for download in download_paths]
    images = [rio.open(image) for image in image_paths]
    composite = combine_images(images)
    cropped = crop_to_box(composite, bbox)
    route = overlay_route(cropped, df[LATITUDE].values, df[LONGITUDE].values, (1, 0, 0))

    '''
    ## First Look
    '''

    #%matplotlib notebook
    matplot_image(gca(), route)

    '''
    ## Calculate Elevation
    '''
    dem = create_elevation(s, cropped)
    elevation = add_elevation(route, dem)
    write_image(elevation, '/tmp/route.tiff')

    '''
    ## Display
    
    The file written above contains 4 layers - RGB and elevation.  
    In theory it should be possible to view the 3D image in a GIS tool like qgis.
    In practice it's close, but not quite there.  Hopefully the next version of qgis will fix the issues. 
    
    Originally I hope that myavi would be able to display this image in the notebook, but it appears to be
    largely unmaintained and bit-rotting - I could not get a complete install 
    (with the necessary GUI toolkit needed for full display).
    '''



from logging import getLogger

from sqlalchemy.orm import sessionmaker

from .database import SystemConstant, Process, MappedDatabase
from .support import SystemBase
from .tables.system import Progress
from ..commands.args import SYSTEM

log = getLogger(__name__)


class System(MappedDatabase):

    def __init__(self, args):
        super().__init__(SYSTEM, SystemConstant, SystemBase, args)
        version = self.get_constant(SystemConstant.DB_VERSION, none=True)
        if version:
            log.info(f'Database version {version}')
        else:
            log.warning('Unconfigured')

    def _sessionmaker(self):
        return sessionmaker(bind=self.engine, expire_on_commit=False)

    def get_constant(self, name, none=False):
        with self.session_context() as s:
            return SystemConstant.get(s, name, none=none)

    def set_constant(self, name, value, force=False):
        with self.session_context() as s:
            return SystemConstant.set(s, name, value, force=force)

    def delete_constant(self, name):
        with self.session_context() as s:
            SystemConstant.delete(s, name)

    def get_process(self, owner, pid):
        with self.session_context() as s:
            return s.query(Process). \
                filter(Process.owner == owner,
                       Process.pid == pid).one()

    def run_process(self, owner, cmd, log_name):
        with self.session_context() as s:
            return Process.run(s, owner, cmd, log_name)  # todo change order

    def delete_process(self, owner, pid, delta_seconds=3):
        with self.session_context() as s:
            Process.delete(s, owner, pid, delta_seconds=delta_seconds)

    def delete_all_processes(self, owner, delta_seconds=3):
        with self.session_context() as s:
            Process.delete_all(s, owner, delta_seconds=delta_seconds)

    def exists_any_process(self, owner):
        with self.session_context() as s:
            return Process.exists_any(s, owner)

    def create_progress(self, name, delta_seconds=3):
        with self.session_context() as s:
            Progress.create(s, name, delta_seconds=delta_seconds)

    def update_progress(self, name, **kargs):
        with self.session_context() as s:
            Progress.update(s, name, **kargs)

    def get_percent(self, name):
        with self.session_context() as s:
            return Progress.get_percent(s, name)

    def wait_for_progress(self, name, timeout=60):
        with self.session_context() as s:
            return Progress.wait_for_progress(s, name, timeout=timeout)




from tests import LogTestCase

from ch2.data import session, std_health_statistics
from ch2.stats.names import REST_HR


class TestHealthBug(LogTestCase):

    def test_health_bug(self):
        '''
        this is only going to work with a database with my data.
        '''
        s = session('-v2')
        df = std_health_statistics(s)
        print(df.describe())
        print(df[REST_HR].describe())



from contextlib import contextmanager
from logging import getLogger
from os import getpid
from sys import argv
from time import sleep, time

from math import floor

from ..commands.args import mm, BASE, VERBOSITY, WORKER, LOG
from ..sql.types import short_cls

log = getLogger(__name__)

DELTA_TIME = 3
SLEEP_TIME = 1
REPORT_TIME = 60


class Workers:

    def __init__(self, system, base, n_parallel, owner, cmd):
        self.__system = system
        self.n_parallel = n_parallel
        self.owner = owner
        self.cmd = cmd
        self.__workers_to_logs = {}  # map from Popen to log index
        self.ch2 = f'{command_root()} {mm(BASE)} {base} {mm(VERBOSITY)} 0'
        self.clear_all()

    def clear_all(self):
        for worker in self.__workers_to_logs:
            log.warning(f'Killing PID {worker.pid} ({worker.args})')
            self.__system.delete_process(self.owner, worker.pid)
        self.__system.delete_all_processes(self.owner)

    def run(self, id, args):
        self.wait(self.n_parallel - 1)
        log_index = self._free_log_index()
        log_name = f'{short_cls(self.owner)}.{log_index}.{LOG}'
        cmd = self.ch2 + f' {mm(LOG)} {log_name} {self.cmd} {mm(WORKER)} {id} {args}'
        worker = self.__system.run_process(self.owner, cmd, log_name)
        self.__workers_to_logs[worker] = log_index

    def wait(self, n_workers=0):
        last_report = 0
        while len(self.__workers_to_logs) > n_workers:
            if time() - last_report > REPORT_TIME:
                log.debug(f'Currently have {len(self.__workers_to_logs)} workers; waiting to drop to {n_workers}')
                last_report = time()
            for worker in list(self.__workers_to_logs.keys()):
                worker.poll()
                process = self.__system.get_process(self.owner, worker.pid)
                if worker.returncode is not None:
                    if worker.returncode:
                        msg = f'Command "{process.command}" exited with return code {worker.returncode} ' + \
                              f'see {process.log} for more info'
                        log.warning(msg)
                        self.clear_all()
                        raise Exception(msg)
                    else:
                        log.debug(f'Command "{process.command}" finished successfully')
                        del self.__workers_to_logs[worker]
                        self.__system.delete_process(self.owner, worker.pid)
            sleep(SLEEP_TIME)
        if last_report:
            log.debug(f'Now have {len(self.__workers_to_logs)} workers')

    def _free_log_index(self):
        used = set(self.__workers_to_logs.values())
        for i in range(self.n_parallel):
            if i not in used:
                return i
        raise Exception('No log available (too many workers)')


def command_root():
    try:
        with open(f'/proc/{getpid()}/cmdline', 'rb') as f:
            line = f.readline()

            def parse():
                word = bytearray()
                for char in line:
                    if char:
                        word.append(char)
                    else:
                        yield word.decode('utf8')
                        word = bytearray()

            words = list(parse())
            if len(argv) > 1:
                i = words.index(argv[1])
                words = words[:i]
            ch2 = ' '.join(words)
            log.debug(f'Using command "{ch2}"')
            return ch2
    except:
        log.warning('Cannot read /proc so assuming that ch2 is started on the command line as "ch2"')
        return 'ch2'


class ProgressTree:

    def __init__(self, size_or_weights, parent=None):
        if isinstance(size_or_weights, int):
            self.__size = size_or_weights
            self.__weights = [1] * self.__size
        else:
            self.__weights = size_or_weights
            self.__size = sum(self.__weights)
        self.__progress = 0
        self.__children = []
        self.__parent = parent
        if parent:
            parent.register(self)

    def _build_weights(self, weights):
        if isinstance(weights, int):
            return [1/weights] * weights

    def register(self, child):
        self.__children.append(child)
        if len(self.__children) > self.__size:
            raise Exception(f'Progress children exceeded size {len(self.__children)}/{self.__size}')

    def local_progress(self):
        if self.__children:
            progress = sum(child.local_progress() * weight
                           for (child, weight) in zip(self.__children, self.__weights)) / self.__size
            return progress
        else:
            return self.__progress / self.__size if self.__size else 1

    def progress(self):
        if self.__parent:
            return self.__parent.progress()
        else:
            return self.local_progress()

    def _log_progress(self):
        local = floor(100 * self.local_progress())
        progress = floor(100 * self.progress())
        log.info(f'Progress: {progress:3d}% (locally {local}%)')

    def increment(self, n=1):
        if self.__children:
            raise Exception('Incrementing a parent node')
        self.__progress += n
        if self.__progress > self.__size:
            raise Exception(f'Progress counter exceeded size {self.__progress}/{self.__size}')
        self._log_progress()

    def complete(self):
        self.__progress = self.__size
        self._log_progress()

    @contextmanager
    def increment_or_complete(self, n=1):
        try:
            yield None
            self.increment(n=n)
        except Exception as e:
            log.debug(f'Completing on {type(e)}: {e}')
            self.complete()
            raise


class SystemProgressTree(ProgressTree):

    def __init__(self, system, name, size_or_weights):
        super().__init__(size_or_weights)
        self.system = system
        self.name = name
        system.create_progress(name)

    def progress(self):
        progress = super().progress()
        self.system.update_progress(self.name, percent=floor(100 * progress))
        return progress





from contextlib import contextmanager
from logging import getLogger
from sqlite3 import OperationalError

from sqlalchemy import create_engine, event, MetaData
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql.functions import count

from . import *
from .support import Base
from ..commands.args import NamespaceWithVariables, NO_OP, make_parser, DATA, DB_EXTN, ACTIVITY
from ..lib.log import make_log_from_args


# mention these so they are "created" (todo - is this needed? missing tables seem to get created anyway)
Source,  Interval, Dummy, Composite, CompositeComponent
ActivityGroup, ActivityJournal, ActivityTimespan, ActivityBookmark
DiaryTopic, DiaryTopicJournal, DiaryTopicField,
ActivityTopic, ActivityTopicJournal, ActivityTopicField,
StatisticName, StatisticJournal, StatisticJournalInteger, StatisticJournalFloat, StatisticJournalText, StatisticMeasure
Segment, SegmentJournal
Pipeline
MonitorJournal
Constant, SystemConstant, Process
ActivitySimilarity, ActivityNearby
Timestamp


log = getLogger(__name__)

# https://stackoverflow.com/questions/13712381/how-to-turn-on-pragma-foreign-keys-on-in-sqlalchemy-migration-script-or-conf
@event.listens_for(Engine, "connect")
def fk_pragma_on_connect(dbapi_con, _con_record):
    cursor = dbapi_con.cursor()

    def pragma(cmd):
        full_cmd = f'PRAGMA {cmd}'
        cursor.execute(full_cmd)

    pragma('foreign_keys=ON;')  # https://www.sqlite.org/pragma.html#pragma_foreign_keys
    pragma('temp_store=MEMORY;')  # https://www.sqlite.org/pragma.html#pragma_temp_store
    pragma('threads=4;')  # https://www.sqlite.org/pragma.html#pragma_threads
    pragma('cache_size=-1000000;')  # 1GB  https://www.sqlite.org/pragma.html#pragma_cache_size
    pragma('secure_delete=OFF;')  # https://www.sqlite.org/pragma.html#pragma_secure_delete
    pragma('journal_mode=WAL;')  # https://www.sqlite.org/wal.html
    pragma(f'busy_timeout={5 * 60 * 1000};')  # https://www.sqlite.org/pragma.html#pragma_busy_timeout
    cursor.close()


@event.listens_for(Engine, 'close')
def analyze_pragma_on_close(dbapi_con, _con_record):
    cursor = dbapi_con.cursor()
    try:
        # this can fail if another process is using the database
        cursor.execute("PRAGMA optimize;")  # https://www.sqlite.org/pragma.html#pragma_optimize
    except OperationalError as e:
        log.debug("Optimize DB aborted (DB likely still in use)")
    finally:
        cursor.close()


class DatabaseBase:

    def __init__(self, path, read_only=False):
        self.path = path
        log.info('Using database at %s' % self.path)
        uri = f'sqlite:///{path}'
        if read_only: uri += '?mode=ro'
        log.debug(f'Connecting to {uri}')
        self.engine = create_engine(uri, echo=False)
        self.session = self._sessionmaker()

    def _sessionmaker(self):
        return sessionmaker(bind=self.engine)

    def no_schema(self, table):
        # https://stackoverflow.com/questions/33053241/sqlalchemy-if-table-does-not-exist
        return not self.engine.dialect.has_table(self.engine, table.__tablename__)

    @contextmanager
    def session_context(self):
        session = self.session()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def __str__(self):
        return f'{self.__class__.__name__} at {self.path}'


class MappedDatabase(DatabaseBase):

    def __init__(self, name, table, base, args, **kargs):
        super().__init__(args.system_path(DATA, name + DB_EXTN), **kargs)
        if self.no_schema(table):
            log.info('Creating tables')
            base.metadata.create_all(self.engine)


class Database(MappedDatabase):

    def __init__(self, args):
        super().__init__(ACTIVITY, Source, Base, args)

    def no_data(self,):
        with self.session_context() as s:
            n_topics = s.query(count(DiaryTopic.id)).scalar()
            n_activities = s.query(count(ActivityGroup.id)).scalar()
            n_statistics = s.query(count(StatisticName.id)).scalar()
            return not (n_topics + n_activities + n_statistics)


def connect(args):
    '''
    Bootstrap from commandline-like args.
    '''
    if len(args) == 1 and isinstance(args[0], str):
        args = args[0].split()
    elif args:
        args = list(args)
    else:
        args = []
    args.append(NO_OP)
    ns = NamespaceWithVariables(make_parser().parse_args(args))
    make_log_from_args(ns)
    db = Database(ns)
    return ns, db


class ReflectedDatabase(DatabaseBase):

    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)
        self.meta = MetaData()
        self.meta.reflect(bind=self.engine)




import re
from logging import getLogger
from os.path import splitext, basename

from pygeotile.point import Point
from sqlalchemy.sql.functions import count

from ..names import LATITUDE, LONGITUDE, M, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, ELEVATION, RAW_ELEVATION, \
    SPORT_GENERIC, COVERAGE, PC, MIN, summaries, AVG, KM
from ..read import MultiProcFitReader, AbortImportButMarkScanned
from ... import FatalException
from ...commands.args import ACTIVITIES, mm, FORCE, DEFAULT, KIT, DEFINE, no
from ...diary.model import TYPE, EDIT
from ...fit.format.records import fix_degrees, merge_duplicates, no_bad_values
from ...fit.profile.profile import read_fit
from ...lib.date import to_time
from ...lib.io import split_fit_path
from ...sql.database import Timestamp, StatisticJournalText
from ...sql.tables.activity import ActivityGroup, ActivityJournal, ActivityTimespan
from ...sql.tables.statistic import StatisticJournalFloat, STATISTIC_JOURNAL_CLASSES, StatisticName, \
    StatisticJournalType, StatisticJournal
from ...sql.tables.topic import ActivityTopicField, ActivityTopic, ActivityTopicJournal
from ...sql.utils import add
from ...srtm.bilinear import bilinear_elevation_from_constant

log = getLogger(__name__)


# duplicate data in
# /home/andrew/archive/fit/batch/DI_CONNECT/DI-Connect-Fitness/UploadedFiles_0-_Part1/andrew@acooke.org_24715592701_tap-sync-18690-cc1dd93225119215a1ea87c584a974ce.fit
# /home/andrew/archive/fit/batch/DI_CONNECT/DI-Connect-Fitness/UploadedFiles_0-_Part1/andrew@acooke.org_24718989709_tap-sync-18690-effaaaffdd06b9419991471bd92d53d5.fit

class ActivityReader(MultiProcFitReader):

    def __init__(self, *args, define=None, sport_to_activity=None, record_to_db=None, kit=True, **kargs):
        from ...commands.upload import ACTIVITY
        self.define = define if define else {}
        self.kit = kit
        self.sport_to_activity = self._assert('sport_to_activity', sport_to_activity)
        self.record_to_db = [(field, name, units, STATISTIC_JOURNAL_CLASSES[type])
                             for field, (name, units, type)
                             in self._assert('record_to_db', record_to_db).items()]
        self.add_elevation = not any(name == ELEVATION for (field, name, units, type) in self.record_to_db)
        self.__ajournal = None  # save for coverage
        super().__init__(*args, sub_dir=ACTIVITY, **kargs)

    def _base_command(self):
        if self.define:
            define = ' '.join(f'{mm(DEFINE)} "{name}={value}"' for name, value in self.define.items()) + ' -- '
        else:
            define = ''
        force = ' ' + mm(FORCE) if self.force else ''
        nokit = ' ' + mm(no(KIT)) if not self.kit else ''
        return f'{ACTIVITIES}{force}{define}{nokit}'

    def _startup(self, s):
        super()._startup(s)
        self.__oracle = bilinear_elevation_from_constant(s)

    def _build_define(self, path):
        define = dict(self.define)
        if self.kit:
            _, kit = split_fit_path(path)
            if kit:
                log.debug(f'Adding {KIT}={kit} to definitions')
                define[KIT] = kit
            else:
                log.debug(f'No kit in {path}')
        return define

    def _read_data(self, s, file_scan):
        log.info('Reading activity data from %s' % file_scan)
        records = self.parse_records(read_fit(file_scan.path))
        define = self._build_define(file_scan.path)
        ajournal, activity_group, first_timestamp = self._create_activity(s, file_scan, define, records)
        return ajournal, (ajournal, activity_group, first_timestamp, file_scan, define, records)

    @staticmethod
    def parse_records(data):
        return ActivityReader.read_fit_file(data, merge_duplicates, fix_degrees, no_bad_values)

    @staticmethod
    def read_sport(path, records):
        try:
            return ActivityReader._first(path, records, 'sport').value.sport.lower()
        except AbortImportButMarkScanned:
            # alternative for some garmin devices (florian)
            return ActivityReader._first(path, records, 'session').value.sport.lower()

    @staticmethod
    def read_first_timestamp(path, records):
        return ActivityReader._first(path, records, 'event', 'record').value.timestamp

    @staticmethod
    def read_last_timestamp(path, records):
        return ActivityReader._last(path, records, 'event', 'record').value.timestamp

    def _create_activity(self, s, file_scan, define, records):
        first_timestamp = self.read_first_timestamp(file_scan.path, records)
        last_timestamp = self.read_last_timestamp(file_scan.path, records)
        log.debug(f'Time range: {first_timestamp.timestamp()} - {last_timestamp.timestamp()}')
        sport = self.read_sport(file_scan.path, records)
        activity_group = self._activity_group(s, file_scan.path, sport, self.sport_to_activity, define)
        log.info(f'{activity_group} from {sport} / {define}')
        if self.force:
            self._delete_journals(s, activity_group, first_timestamp, last_timestamp, file_scan)
        else:
            self._check_journals(s, activity_group, first_timestamp, last_timestamp, file_scan)
        ajournal = add(s, ActivityJournal(activity_group=activity_group,
                                          start=first_timestamp, finish=first_timestamp,  # will be over-written later
                                          file_hash_id=file_scan.file_hash_id))
        return ajournal, activity_group, first_timestamp

    def _activity_group(self, s, path, sport, lookup, define):
        log.debug(f'Current lookup: {lookup}; sport: {sport}; define: {define}')
        if isinstance(lookup, str):
            return self._lookup_activity_group(s, lookup)
        if sport in lookup:
            return self._activity_group(s, path, sport, lookup[sport], define)
        for key, values in define.items():
            for value in values.split(','):
                if key in lookup and value in lookup[key]:
                    return self._activity_group(s, path, sport, lookup[key][value], define)
        if DEFAULT in lookup:
            return self._activity_group(s, path, sport, lookup[DEFAULT], define)
        log.warning('Unrecognised sport: "%s" in %s' % (sport, path))
        if sport in (SPORT_GENERIC,):
            raise Exception(f'Ignoring {sport} entry')
        else:
            raise FatalException(f'There is no group configured for {sport} entries in the FIT file.')

    def _lookup_activity_group(self, s, name):
        activity_group = ActivityGroup.from_name(s, name, optional=True)
        if not activity_group:
            activities = s.query(ActivityGroup).all()
            if activities:
                log.info('Available activity group:')
                for activity_group in activities:
                    log.info('%s - %s' % (activity_group.name, activity_group.description))
            else:
                log.error('No activity groups defined - configure system correctly')
            raise Exception('ActivityGroup "%s" is not defined' % name)
        return activity_group

    def _overlapping_journals(self, s, what, activity_group, first_timestamp, last_timestamp):
        return s.query(what). \
                filter(ActivityJournal.activity_group == activity_group,
                       ActivityJournal.start < last_timestamp,
                       ActivityJournal.finish >= first_timestamp)

    def _file_hash_journals(self, s, what, file_scan):
        return s.query(what). \
                filter(ActivityJournal.file_hash == file_scan.file_hash)

    def _delete_query(self, s, query):
        for journal in query.all():
            Timestamp.clear(s, owner=self.owner_out, source=journal)
            log.debug(f'Deleting {journal}')
            s.delete(journal)

    def _delete_journals(self, s, activity_group, first_timestamp, last_timestamp, file_scan):
        log.debug('Deleting overlapping journals')
        self._delete_query(s, self._overlapping_journals(s, ActivityJournal, activity_group,
                                                         first_timestamp, last_timestamp))
        log.debug('Deleting journals from same file')
        self._delete_query(s, self._file_hash_journals(s, ActivityJournal, file_scan))
        s.commit()

    def _check_journals(self, s, activity_group, first_timestamp, last_timestamp, file_scan):
        log.debug('Checking for overlapping journals')
        if self._overlapping_journals(s, count(ActivityJournal.id), activity_group,
                                      first_timestamp, last_timestamp).scalar():
            log.warning(f'Overlapping activities for {first_timestamp} - {last_timestamp}')
            raise AbortImportButMarkScanned()
        log.debug('Checking for previous file')
        if self._file_hash_journals(s, count(ActivityJournal.id), file_scan).scalar():
            log.warning(f'Repeated scan of file {file_scan.path}')
            raise Exception(file_scan.path)  # this should not happen since hash is checked before import

    def _load_define(self, s, define, ajournal):
        for name, value in define.items():
            log.debug(f'Setting {name} = {value}')
            StatisticJournalText.add(s, name, None, None, self.owner_out, ajournal.activity_group,
                                     ajournal, value, ajournal.start)

    @staticmethod
    def _save_name(s, ajournal, file_scan):
        from ...config import add_activity_topic_field
        log.debug('Saving name')
        # first, do we have the 'Name' field defined for all activities?
        # this should be triggered at most once if it was not already defined
        if not s.query(ActivityTopicField). \
                join(StatisticName). \
                filter(StatisticName.name == ActivityTopicField.NAME,
                       StatisticName.constraint == ajournal.activity_group,
                       StatisticName.owner == ActivityTopic,
                       ActivityTopicField.activity_topic_id == None).one_or_none():
            add_activity_topic_field(s, None, ActivityTopicField.NAME, -10, StatisticJournalType.TEXT,
                                     ajournal.activity_group, model={TYPE: EDIT})
        # second, do we already have a journal for this file, or do we need to add one?
        source = s.query(ActivityTopicJournal). \
            filter(ActivityTopicJournal.file_hash_id == file_scan.file_hash_id). \
            one_or_none()
        if not source:
            source = add(s, ActivityTopicJournal(file_hash_id=file_scan.file_hash_id))
        # and third, does that journal contain a name?
        if not s.query(StatisticJournal). \
                join(StatisticName). \
                filter(StatisticJournal.source == source,
                       StatisticName.owner == ActivityTopic,
                       StatisticName.constraint == ajournal.activity_group,
                       StatisticName.name == ActivityTopicField.NAME). \
                one_or_none():
            value = splitext(basename(file_scan.path))[0]
            StatisticJournalText.add(s, ActivityTopicField.NAME, None, None, ActivityTopic, ajournal.activity_group,
                                     source, value, ajournal.start)

    def _load_data(self, s, loader, data):

        ajournal, activity_group, first_timestamp, file_scan, define, records = data
        timespan, warned, logged, last_timestamp = None, 0, 0, to_time(0.0)
        self._load_define(s, define, ajournal)
        self._save_name(s, ajournal, file_scan)
        self.__ajournal = ajournal

        log.debug(f'Loading {self.record_to_db}')

        def is_event(record, *types):
            event = False
            if record.name == 'event':
                event = record.value.event == 'timer' and record.value.event_type in types
                if event: log.debug(f'{types} at {record.timestamp}')
            return event

        have_timespan = any(is_event(record, 'start') for record in records)
        only_records = list(filter(lambda x: x.name == 'record', records))
        final_timestamp = only_records[-1].timestamp
        if not have_timespan:
            first_timestamp = only_records[0].timestamp
            log.warning('Experimental handling of data without timespans')
            timespan = add(s, ActivityTimespan(activity_journal=ajournal, start=first_timestamp))

        for record in records:

            if have_timespan and is_event(record, 'start'):
                if timespan:
                    log.warning('Ignoring start with no corresponding stop (possible lost data?)')
                else:
                    timespan = add(s, ActivityTimespan(activity_journal=ajournal,
                                                       start=record.value.timestamp,
                                                       finish=record.value.timestamp))

            elif record.name == 'record':
                if record.value.timestamp > last_timestamp:
                    lat, lon, timestamp = None, None, record.value.timestamp
                    # customizable loader
                    for field, name, units, type in self.record_to_db:
                        value = record.data.get(field, None)
                        if logged < 3:
                            log.debug(f'{name} = {value}')
                        if value is not None:
                            value = value[0][0]
                            if units == KM:  # internally everything uses M
                                value /= 1000
                            loader.add(name, units, None, activity_group, ajournal, value, timestamp, type)
                            if name == LATITUDE:
                                lat = value
                            elif name == LONGITUDE:
                                lon = value
                    logged += 1
                    # values derived from lat/lon
                    if lat is not None and lon is not None:
                        x, y = Point.from_latitude_longitude(lat, lon).meters
                        loader.add(SPHERICAL_MERCATOR_X, M, None, activity_group, ajournal, x, timestamp,
                                   StatisticJournalFloat)
                        loader.add(SPHERICAL_MERCATOR_Y, M, None, activity_group, ajournal, y, timestamp,
                                   StatisticJournalFloat)
                        if self.add_elevation:
                            elevation = self.__oracle.elevation(lat, lon)
                            if elevation:
                                loader.add(RAW_ELEVATION, M, None, activity_group, ajournal, elevation, timestamp,
                                           StatisticJournalFloat)
                else:
                    log.warning('Ignoring duplicate record data for %s at %s - some data may be missing' %
                                (file_scan.path, record.value.timestamp))
                last_timestamp = record.value.timestamp
                if not have_timespan:
                    ajournal.finish = record.value.timestamp

            elif have_timespan and is_event(record, 'stop_all', 'stop'):
                if timespan:
                    timespan.finish = record.value.timestamp
                    ajournal.finish = record.value.timestamp
                    timespan = None
                else:
                    log.debug('Ignoring stop with no corresponding start (possible lost data?)')

        if timespan:
            log.warning('Cleaning up dangling timespan')
            timespan.finish = final_timestamp

    def _read(self, s, path):
        loader = super()._read(s, path)
        for (name, constraint), percent in loader.coverage_percentages():
            StatisticJournalFloat.add(s, COVERAGE, PC, summaries(MIN, AVG), self.owner_out,
                                      f'{name} / {constraint.name}', self.__ajournal, percent,
                                      self.__ajournal.start)
        s.commit()



from tests import LogTestCase

from sqlalchemy import Column, Integer, ForeignKey, create_engine
from sqlalchemy.event import listen
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy.sql.functions import count

from ch2.sql.support import Base


class Database:

    def __init__(self):
        self.engine = create_engine('sqlite:///:memory:', echo=True)
        listen(self.engine, 'connect', self.fk_pragma_on_connect)
        self.session = sessionmaker(bind=self.engine)
        Base.metadata.create_all(self.engine)

    @staticmethod
    def fk_pragma_on_connect(dbapi_con, con_record):
        dbapi_con.execute('pragma foreign_keys=ON')


class Parent(Base):

    __tablename__ = 'parent'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False)
    parent_ref_id = Column(Integer, ForeignKey('other.id', ondelete='cascade'))
    parent_ref = relationship('Other')

    __mapper_args__ = {
        'polymorphic_identity': 0,
        'polymorphic_on': type
    }


class Child(Parent):

    __tablename__ = 'child'

    id = Column(Integer, ForeignKey('parent.id', ondelete='cascade'), primary_key=True)
    child_ref_id = Column(Integer, ForeignKey('other.id', ondelete='cascade'))
    child_ref = relationship('Other')

    __mapper_args__ = {
        'polymorphic_identity': 1,
        'passive_deletes': True
    }


class Other(Base):

    __tablename__ = 'other'

    id = Column(Integer, primary_key=True)


class TestInheritance(LogTestCase):

    def database(self):
        db = Database()
        s = db.session()
        o1 = Other()
        s.add(o1)
        o2 = Other()
        s.add(o2)
        s.add(Child(parent_ref=o1, child_ref=o2))
        s.commit()
        return db

    def test_delete_child_instance(self):
        s = self.database().session()
        c = s.query(Child).one()
        s.delete(c)
        self.assertEqual(s.query(count(Parent.id)).scalar(), 0)

    def test_delete_child_sql(self):
        s = self.database().session()
        s.query(Child).delete()
        # !!! this is because the ondelete cascade goes "the other way"
        self.assertEqual(s.query(count(Parent.id)).scalar(), 1)

    def test_delete_parent_instance(self):
        s = self.database().session()
        p = s.query(Parent).one()
        s.delete(p)
        self.assertEqual(s.query(count(Child.id)).scalar(), 0)

    def test_delete_parent_sql(self):
        s = self.database().session()
        s.query(Parent).delete()
        self.assertEqual(s.query(count(Child.id)).scalar(), 0)

    def test_delete_child_ref_instance(self):
        db = self.database()
        s = db.session()
        o = s.query(Child).one().child_ref
        s.delete(o)
        s.commit()
        s = db.session()
        # !!! this is because SQL deletes the child via cascade, but not the parent (see test_delete_child_sql)
        # (we don't have a backref on Other, so SQLAlchemy isn't going to do anything - have no idea if it would)
        self.assertEqual(s.query(count(Parent.id)).scalar(), 1)
        self.assertEqual(s.query(count(Child.id)).scalar(), 0)

    def test_delete_parent_ref_instance(self):
        db = self.database()
        s = db.session()
        o = s.query(Child).one().parent_ref
        s.delete(o)
        s.commit()
        s = db.session()
        self.assertEqual(s.query(count(Parent.id)).scalar(), 0)
        self.assertEqual(s.query(count(Child.id)).scalar(), 0)




from logging import getLogger

from ..commands.args import PATH, FORCE, FAST, WORKER, parse_pairs, KARG, BASE
from ..sql.tables.pipeline import PipelineType
from ..stats.pipeline import run_pipeline

log = getLogger(__name__)


def monitor(args, sys, db):
    '''
## monitor

    > ch2 monitor PATH [PATH ...]

Read monitor data from FIT files.

Note: When using bash use `shopt -s globstar` to enable ** globbing.
    '''
    run_monitor_pipelines(sys, db, args[BASE], paths=args[PATH], force=args[FORCE],
                          worker=args[WORKER] is not None, id=args[WORKER], **parse_pairs(args[KARG]))


def run_monitor_pipelines(sys, db, base, paths=None, force=False, worker=False, id=None, **kargs):
    if paths is None: paths = []
    run_pipeline(sys, db, base, PipelineType.MONITOR, paths=paths, force=force, worker=worker, id=id, **kargs)



from json import dumps
from logging import getLogger

from . import name_constant, add_enum_constant, set_constant
from ..data.climb import Climb, MAX_CLIMB_GRADIENT, MIN_CLIMB_GRADIENT, MAX_CLIMB_REVERSAL, \
    MIN_CLIMB_ELEVATION

log = getLogger(__name__)

CLIMB_CNAME = 'Climb'


def add_climb(s, phi=0.7):
    '''
    Add the constants necessary to auto-detect climbs.
    '''
    log.debug('Adding climb parameters')
    add_enum_constant(s, CLIMB_CNAME, Climb,
                      {'phi': phi,
                       'min_gradient': MIN_CLIMB_GRADIENT,
                       'max_gradient': MAX_CLIMB_GRADIENT,
                       'min_elevation': MIN_CLIMB_ELEVATION,
                       'max_reversal': MAX_CLIMB_REVERSAL},
                      single=True, description='''
Parameters used by climb detection.
* Phi is an exponent used to weight the relative steepness of candidates and should be around 1 (typical range 0.5 to 2).
* Min_gradient is the minimum gradient for a climb (units %).
* Min_elevation is the minimum elevation for a climb (units m) (this, together with min_gradient, implies a minimum distance).
* Max_reversal is the fractional descent allowed within a climb (if it has a value of 0.5 then you can descend half the climbed distance).
                              ''')



from json import dumps
from logging import getLogger

from werkzeug import Response


log = getLogger(__name__)


class JsonResponse(Response):

    default_mimetype = 'application/json'

    def __init__(self, content, **kargs):
        super().__init__(dumps(content), **kargs)




from logging import getLogger

from ..commands.args import DATE, NAME, VALUE, DELETE, FORCE, mm, COMMAND, CONSTANTS, SET, SUB_COMMAND, ADD, \
    SHOW, REMOVE, CONSTRAINT, DESCRIPTION, SINGLE, VALIDATE
from ..sql.tables.constant import Constant, ValidateNamedTuple
from ..sql.tables.statistic import StatisticJournal, StatisticName, StatisticJournalType
from ..sql.types import lookup_cls
from ..sql.utils import add

log = getLogger(__name__)


def constants(args, system, db):
    '''
## constants

    > ch2 constants show [NAME [DATE]]

Lists constants to stdout.

    > ch2 constants add NAME

Defines a new constant.

    > ch2 constants set NAME VALUE [DATE]

Adds an entry for the constant.  If date is omitted a single value is used for all time
(so any previously defined values are deleted).

Note that adding / removing constants (ie their names) is separate from setting / deleting entries (ie their values).

    > ch2 constants delete NAME [DATE]

Deletes an entry.

    > ch2 constants remove NAME

Remove a constant (the associated entries must have been deleted first).

Names can be matched by SQL patterns.  So FTHR.% matches both FTHR.Run and FTHR.Bike, for example.
In such a case "entry" in the descriptions above may refer to multiple entries.

TODO - Constraint handling is confused and confusing.
    '''
    name, cmd = args[NAME], args[SUB_COMMAND]
    with db.session_context() as s:
        if cmd == ADD:
            add_constant(s, name, constraint=args[CONSTRAINT], description=args[DESCRIPTION],
                         single=args[SINGLE], validate=args[VALIDATE])
        else:
            if name:
                constants = constants_like(s, name)
                if not constants:
                    raise Exception(f'Name "{name}" matched no entries (see `{COMMAND} {CONSTANTS}`)')
            else:
                constants = []
            if cmd == REMOVE:
                if len(constants) > 1 and not args[FORCE]:
                    raise Exception(f'Use {mm(FORCE)} to remove multiple constants')
                remove_constants(s, constants)
            else:
                date = args[DATE]
                if cmd == SET:
                    set_constants(s, constants, date, args[VALUE], args[FORCE])
                elif cmd == DELETE:
                    if not date and not args[FORCE]:
                        raise Exception(f'Use {mm(FORCE)} to delete all entries for a Constant')
                    delete_constants(s, constants, date)
                elif cmd == SHOW:
                    print_constants(s, constants, name, date)


def constants_like(s, name):
    constants = s.query(Constant).filter(Constant.name.like(name)).order_by(Constant.name).all()
    if not constants:
        all_constants = s.query(Constant).all()
        if all_constants:
            log.info('Available constants:')
            for constants in all_constants:
                log.info('%s - %s' % (constants.statistic_name.name, constants.statistic_name.description))
        else:
            log.error('No constants defined - configure system correctly')
        raise Exception('Constant "%s" is not defined' % name)
    return constants


def add_constant(s, name, constraint=None, description=None, single=False, validate=None):
    if s.query(StatisticName). \
            filter(StatisticName.name == name,
                   StatisticName.owner == Constant,
                   StatisticName.constraint == constraint).one_or_none():
        raise Exception(f'Constant {name} (constraint {constraint}) already exists')
    validate_cls, validate_args, validate_kargs = None, [], {}
    if validate:
        lookup_cls(validate)
        validate_cls, validate_kargs = ValidateNamedTuple, {'tuple_cls': validate}
    statistic_name = add(s, StatisticName(name=name, owner=Constant, constraint=constraint,
                                          units=None, description=description,
                                          statistic_journal_type=StatisticJournalType.TEXT))
    add(s, Constant(statistic_name=statistic_name, name=name, single=single,
                    validate_cls=validate_cls, validate_args=validate_args, validate_kargs=validate_kargs))
    log.info(f'Added {name}')


def set_constants(s, constants, date, value, force):
    if not date:
        log.info('Checking any previous values')
        journals = []
        for constant in constants:
            journals += s.query(StatisticJournal).join(StatisticName, Constant). \
                filter(Constant.id == constant.id).all()
        if journals:
            log.info(f'Need to delete {len(journals)} ConstantJournal entries')
            if not force:
                raise Exception(f'Use {mm(FORCE)} to confirm deletion of prior values')
            for journal in journals:
                s.delete(journal)
    for constant in constants:
        constant.add_value(s, value, date=date)
        log.info(f'Added value {value} at {date} for {constant.name}')
    log.warning('You may want to (re-)calculate statistics')


def delete_constants(s, constants, date):
    if date:
        for constant in constants:
            for repeat in range(2):
                journal = s.query(StatisticJournal).join(StatisticName, Constant). \
                    filter(Constant.id == constant.id,
                           StatisticJournal.time == date).one_or_none()
                if repeat:
                    log.info(f'Deleting {constant.name} on {journal.time}')
                    s.delete(journal)
                else:
                    if not journal:
                        raise Exception(f'No entry for {constant.name} on {date}')
    else:
        for constant in constants:
            for journal in s.query(StatisticJournal).join(StatisticName, Constant). \
                    filter(Constant.id == constant.id).order_by(StatisticJournal.time).all():
                log.info(f'Deleting {constant.name} on {journal.time}')
                s.delete(journal)


def print_constants(s, constants, name, date):
    if not constants:
        constants = s.query(Constant).order_by(Constant.name).all()
        if not constants:
            raise Exception('No Constants defined - configure system')
    print()
    for constant in constants:
        if not date:
            description = constant.statistic_name.description \
                if constant.statistic_name.description else '[no description]'
            print(f'{constant.name}: {description}')
            if name:  # only print values if we're not listing all
                found = False
                for journal in s.query(StatisticJournal).join(StatisticName, Constant). \
                        filter(Constant.id == constant.id).order_by(StatisticJournal.time).all():
                    print(f'{journal.time}: {journal.value}')
                    found = True
                if not found:
                    log.warning(f'No values found for {constant.name}')
            print()
        else:
            journal = constant.at(s, date=date)
            if journal:
                print(f'{constant.name} {journal.source.time} {journal.value}')
            else:
                log.warning(f'No values found for {constant.name}')


def remove_constants(s, constants):
    for do in (0, 1):
        for constant in constants:
            if do:
                log.warning(f'Deleting {constant.name}')
                s.query(StatisticName).filter(StatisticName.id == constant.statistic_name_id).delete()
                s.query(Constant).filter(Constant.id == constant.id).delete()
            else:
                if s.query(StatisticJournal). \
                        join(StatisticName, Constant). \
                        filter(Constant.id == constant.id).count():
                    raise Exception(f'Values still defined for {constant.name}')




from logging import getLogger
from types import SimpleNamespace

from sqlalchemy.orm.exc import NoResultFound

from . import journal_imported, match_statistic_name, copy_statistic_journal
from ...sql import DiaryTopic, DiaryTopicJournal
from ...lib.log import log_current_exception

log = getLogger(__name__)


def upgrade_diary(record, old, new):
    if not diary_imported(record, new):
        log.debug(f'Trying to copy diary topic data from {old} to {new}')
        with old.session_context() as old_s:
            diary_topic = old.meta.tables['diary_topic']
            for old_diary_topic in old_s.query(diary_topic).filter(diary_topic.c.parent_id == None).all():
                log.info(f'Found old (root) diary_topic {old_diary_topic}')
                copy_diary_topic_fields(record, old_s, old, old_diary_topic, new)


def diary_imported(record, new):
    return journal_imported(record, new, DiaryTopicJournal, 'Diary')


def copy_diary_topic_fields(record, old_s, old, old_diary_topic, new):
    try:
        with new.session_context() as new_s:
            try:
                new_diary_topic = new_s.query(DiaryTopic). \
                    filter(DiaryTopic.name == old_diary_topic.name,
                           DiaryTopic.schedule == old_diary_topic.schedule).one()
            except NoResultFound:
                record.raise_(f'No new equivalent to diary topic {old_diary_topic.name} '
                              f'(schedule {old_diary_topic.schedule})')
            constraint = str(new_diary_topic)
        copy_diary_topic_fields_with_constraint(record, old_s, old, old_diary_topic, new, constraint)
    except:
        # do we want to generalize this?  also, is there a neater way of mutating a copy?
        log_current_exception()


def copy_diary_topic_fields_with_constraint(record, old_s, old, old_diary_topic, new, constraint):
    # this ignores the schedule on the diary_topic_field because it only copies statistics
    # i think this is ok?
    log.debug(f'Trying to copy diary_topic_fields for diary_topic {old_diary_topic}')
    diary_topic_field = old.meta.tables['diary_topic_field']
    for old_diary_topic_field in old_s.query(diary_topic_field). \
            filter(diary_topic_field.c.diary_topic_id ==
                   (old_diary_topic.id if old_diary_topic else None)).all():
        log.debug(f'Found old diary_topic_field {old_diary_topic_field}')
        try:
            statistic_name = old.meta.tables['statistic_name']
            old_statistic_name = old_s.query(statistic_name). \
                filter(statistic_name.c.id == old_diary_topic_field.statistic_name_id).one()
            log.debug(f'Found old statistic_name {old_statistic_name}')
            with new.session_context() as new_s:
                new_statistic_name = match_statistic_name(record, old_statistic_name, new_s, DiaryTopic, constraint)
                copy_diary_topic_journal_entries(record, old_s, old, old_statistic_name, new_s, new_statistic_name)
        except:
            log_current_exception()
    parent_id = old_diary_topic.id
    diary_topic = old.meta.tables['diary_topic']
    for old_diary_topic in old_s.query(diary_topic).filter(diary_topic.c.parent_id == parent_id).all():
        log.info(f'Found old diary_topic {old_diary_topic}')
        copy_diary_topic_fields(record, old_s, old, old_diary_topic, new)


def copy_diary_topic_journal_entries(record, old_s, old, old_statistic_name, new_s, new_statistic_name):
    log.debug(f'Trying to find statistic_journal entries for {old_statistic_name}')
    statistic_journal = old.meta.tables['statistic_journal']
    diary_topic_journal = old.meta.tables['diary_topic_journal']
    for old_statistic_journal in old_s.query(statistic_journal). \
            join(diary_topic_journal, statistic_journal.c.source_id == diary_topic_journal.c.id). \
            filter(statistic_journal.c.statistic_name_id == old_statistic_name.id).all():
        log.debug(f'Found old statistic_journal {old_statistic_journal}')
        old_diary_topic_journal = old_s.query(diary_topic_journal). \
            filter(diary_topic_journal.c.id == old_statistic_journal.source_id).one()
        log.debug(f'Found old diary_topic_journal {old_diary_topic_journal}')
        new_diary_topic_journal = DiaryTopicJournal.get_or_add(new_s, old_diary_topic_journal.date)
        log.debug(f'Found new diary_topic_journal {new_diary_topic_journal}')
        copy_statistic_journal(record, old_s, old, old_statistic_name, old_statistic_journal,
                               new_s, new_statistic_name, new_diary_topic_journal)






from logging import getLogger
import datetime as dt

from sqlalchemy import desc

from ..display import Reader
from ...diary.model import optional_text, text
from ...lib import local_date_to_time
from ...sql import ActivityGroup, ActivityJournal, Achievement

log = getLogger(__name__)


class AchievementDiary(Reader):

    @optional_text('Achievements')
    def _read_date(self, s, date):
        journals = list(self._journals_at_date(s, date))
        for journal in journals:
            achievements = list(self._achievements_for_journal(s, journal))
            if not achievements: achievements = [text('Get out and try harder!')]
            if len(journals) > 1:
                achievements = [text('journal name')] + achievements
                yield achievements
            else:
                yield from achievements

    def _journals_at_date(self, s, date):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        for activity_group in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            for ajournal in s.query(ActivityJournal). \
                    filter(ActivityJournal.finish >= start,
                           ActivityJournal.start < finish,
                           ActivityJournal.activity_group == activity_group). \
                    order_by(ActivityJournal.start).all():
                yield ajournal

    def _achievements_for_journal(self, s, journal):
        for achievement in s.query(Achievement). \
                filter(Achievement.activity_journal == journal). \
                order_by(desc(Achievement.sort)).limit(5).all():   # limit 5 for compact display
            yield text(achievement.text)

    def _read_schedule(self, s, date, schedule):
        return
        yield




from colorsys import hsv_to_rgb

from bokeh.models import PanTool, ZoomInTool, ZoomOutTool, ResetTool, HoverTool, Range1d

from ...stats.names import DISTANCE_KM, LOCAL_TIME


def tooltip(name):
    return name, '@{%s}' % name


def make_tools(y):
    tools = [PanTool(dimensions='width'),
             ZoomInTool(dimensions='width'), ZoomOutTool(dimensions='width'),
             ResetTool(),
             HoverTool(tooltips=[tooltip(x) for x in (y, DISTANCE_KM, LOCAL_TIME)])]
    return tools


def make_range(column, lo=None, hi=None):
    column = column.dropna()
    mn, mx = column.min() if lo is None else lo, column.max() if hi is None else hi
    delta = mx - mn
    return Range1d(start=mn - 0.1 * delta, end=mx + 0.1 * delta)


def evenly_spaced_hues(n, saturation=1, value=1, stagger=1):
    for i in range(n):
        r, g, b = [int(x * 255) for x in hsv_to_rgb(((i * stagger) % n) / n, saturation, value)]
        yield f'#{r:02x}{g:02x}{b:02x}'.upper()


def get_renderer(f, name):
    for renderer in f.renderers:
        if renderer.name == name:
            return renderer
    raise Exception(f'No rendered {name} in {f}')




from glob import glob
from logging import getLogger
from os import makedirs, unlink
from os.path import basename, join, exists, dirname

from math import sqrt

from .activities import run_activity_pipelines
from .garmin import run_garmin
from .monitor import run_monitor_pipelines
from .statistics import run_statistic_pipelines
from ..commands.args import KIT, FAST, UPLOAD, BASE, FORCE, UNSAFE, DELETE, PATH, REPLACE, mm, base_system_path, \
    PERMANENT
from ..diary.model import TYPE
from ..lib.date import time_to_local_time, Y, YMDTHMS
from ..lib.io import data_hash, split_fit_path, touch
from ..lib.log import log_current_exception, Record
from ..lib.utils import clean_path, slow_warning
from ..lib.workers import ProgressTree, SystemProgressTree
from ..sql import KitItem, FileHash, ActivityJournal, Source, FileScan
from ..stats.names import TIME
from ..stats.read import AbortImportButMarkScanned
from ..stats.read.activity import ActivityReader
from ..stats.read.monitor import MonitorReader

log = getLogger(__name__)


STREAM = 'stream'
DATA = 'data'
EXTRA = 'extra'
HASH = 'hash'
NAME = 'name'
ACTIVITY = 'activity'
MONITOR = 'monitor'
SPORT = 'sport'
DIR = 'dir'
DOT_FIT = '.fit'
READ_PATH = 'read-path'
WRITE_PATH = 'write-path'
REPLACES = 'replaces'


def upload(args, system, db):
    '''
## upload

    > ch2 upload --kit ITEM [ITEM...] -- PATH [PATH ...]

Upload FIT files, storing the data in the permanent store (Data.Dir) on file system.
Optionally (if not --fast), scan the files and appropriate entries to the database.
Both monitor and activity files are accepted.

Files are checked for duplication on uploading (before being scanned).

If the uploaded file is mapped to a file path that already exists then we check the following cases:
* If the hash matches then the new data are discarded (duplicate).
* If the hash is different, it is an error (to avoid losing activity diary entries which are keyed by hash).

If the uploaded file has a hash that matches a file already read into the database, but the file path does not match,
then it is an error (internal error with inconsistent naming or duplicate data).

If --unsafe is given then files whose hashes match existing files are simply ignored.
This allows data downloaded en massed from Garmin to be uploaded ignoring already-uploaded data (with kit).

In short, hashes should map 1-to-1 to file names and to activities, and it is an error if they do not.

### Examples

    > ch2 upload --kit cotic -- ~/fit/2018-01-01.fit

will store the given file, add activity data to the database (associated with the kit 'cotic'), check for
new monitor data, and update statistics.

Note: When using bash use `shopt -s globstar` to enable ** globbing.
    '''
    record = Record(log)
    nfiles, files = open_files(args[PATH])
    upload_files_and_update(record, system, db, args[BASE], files=files, nfiles=nfiles, force=args[FORCE],
                            items=args[KIT], fast=args[FAST], unsafe=args[UNSAFE], delete=args[DELETE],
                            replace=args[REPLACE])


class SkipFile(Exception):
    pass


def open_files(paths):
    # converts a list of paths to a map of file names to file dicts with open streams
    n = len(paths)

    def files():
        # use an iterator here to avoid opening too many files at once
        for path in paths:
            path = clean_path(path)
            name = basename(path)
            log.debug(f'Reading {path}')
            stream = open(path, 'rb')
            yield {STREAM: stream, NAME: name, READ_PATH: path}

    return n, files()


def check_items(s, items):
    for item in items:
        if not s.query(KitItem).filter(KitItem.name == item).one_or_none():
            raise Exception(f'Kit item {item} does not exist')


def read_file(file):
    # add DATA to dicts with STREAM
    log.debug(f'Reading {file[NAME]}')
    file[DATA] = file[STREAM].read()
    file[STREAM].close()


def hash_file(file):
    # add HASH to dicts with DATA
    log.debug(f'Hashing {file[NAME]}')
    file[HASH] = data_hash(file[DATA])
    log.debug(f'Hash of {file[NAME]} is {file[HASH]}')


def check_path(file, unsafe=False):
    path, name = file[WRITE_PATH], file[NAME]
    gpath, _ = split_fit_path(path)
    match = glob(gpath)
    if match:
        path2 = clean_path(match[0])
        log.debug(f'A file already exists at {path2}')
        with open(path2, 'rb') as input:
            hash = data_hash(input.read())
            if REPLACES in file and path2 in file[REPLACES]:
                log.warning(f'Ignoring conflict at {path2} because {mm(REPLACE)}')
            elif hash == file[HASH] and path == path2:
                # touch in case we deleted the activity and need to read again
                touch(path)
                raise SkipFile(f'Duplicate file {name} at {path2}')
            else:
                msg = f'File {name} for {path} does not match the file already at {path2} (different hash or kit)'
                if unsafe:
                    raise SkipFile(msg)
                else:
                    raise Exception(msg)
    log.debug(f'File {name} cleared for path {path}')


def check_hash(s, file, unsafe):
    hash, path, name = file[HASH], file[WRITE_PATH], file[NAME]
    file_hash = s.query(FileHash).filter(FileHash.hash == hash).one_or_none()
    if file_hash and file_hash.file_scan:
        log.debug(f'A file was already scanned with hash {hash}')
        scanned_path = file_hash.file_scan.path
        if scanned_path == path:
            # touch in case we deleted the activity and need to read again
            touch(path)
            # should never happen because would be caught in check_path
            raise SkipFile(f'Duplicate file {name} at {path}')
        elif REPLACES in file and scanned_path in file[REPLACES]:
            log.warning(f'Ignoring conflict at {scanned_path} because {mm(REPLACE)}')
        else:
            msg = f'File {name} for {path} matches the file already at {scanned_path} (same hash)'
            if unsafe:
                raise SkipFile(msg)
            else:
                raise Exception(msg)
    log.debug(f'File {name} cleared for hash {hash}')


def check_file(s, file, unsafe=False):
    check_path(file, unsafe=unsafe)
    check_hash(s, file, unsafe=unsafe)


def parse_fit_data(file, items=None):
    try:
        # add TIME and TYPE and EXTRA (and maybe SPORT) given (fit) DATA and NAME
        try:
            records = MonitorReader.parse_records(file[DATA])
            file[TIME] = MonitorReader.read_first_timestamp(file[NAME], records)
            file[TYPE] = MONITOR
            file[EXTRA] = ':' + file[HASH][0:5]
            log.debug(f'File {file[NAME]} contains monitor data')
        except AbortImportButMarkScanned:
            records = ActivityReader.parse_records(file[DATA])
            file[TIME] = ActivityReader.read_first_timestamp(file[NAME], records)
            file[SPORT] = ActivityReader.read_sport(file[NAME], records)
            file[TYPE] = ACTIVITY
            file[EXTRA] = '-' + ','.join(items) if items else ''
            log.debug(f'File {file[NAME]} contains activity data')
    except Exception as e:
        log_current_exception(traceback=False)
        raise Exception(f'Could not parse {file[NAME]} as a fit file')


def build_path(data_dir, file):
    # add PATH given TIME, TYPE, SPORT and EXTRA
    date = time_to_local_time(file[TIME], fmt=YMDTHMS)
    year = time_to_local_time(file[TIME], fmt=Y)
    file[DIR] = join(data_dir, file[TYPE], year)
    if SPORT in file: file[DIR] = join(file[DIR], file[SPORT])
    file[WRITE_PATH] = join(file[DIR], date + file[EXTRA] + DOT_FIT)
    log.debug(f'Target directory is {file[DIR]}')


def remove_previous_activity(s, file):
    activity_journal = s.query(ActivityJournal).filter(ActivityJournal.start == file[TIME]).one_or_none()
    if not activity_journal:
        log.warning(f'No previous activity matching {file[NAME]}')
    else:
        # delete after reading (could be same as READ_PATH and we might fail to write)
        file[REPLACES] = [clean_path(row[0]) for row in
                          s.query(FileScan.path).join(FileHash).
                              filter(FileHash.id == activity_journal.file_hash_id).all()]
        log.debug(f'Found replacements {file[REPLACES]}')
        source = s.query(Source).filter(Source.id == activity_journal.id).one()
        slow_warning(f'Deleting activity at {time_to_local_time(activity_journal.start)}')
        s.delete(source)
        s.commit()


def remove_previous_files(file):
    if REPLACES in file:
        for path in file[REPLACES]:
            if path == file[WRITE_PATH]:
                log.warning(f'Not deleting {path} since wrote to same location ({mm(REPLACE)})')
            else:
                slow_warning(f'Deleting {path}')
                unlink(path)


def write_file(file):
    try:
        dir = dirname(file[WRITE_PATH])
        if not exists(dir):
            log.debug(f'Creating {dir}')
            makedirs(dir)
        if exists(file[WRITE_PATH]):
            log.warning(f'Overwriting data at {file[WRITE_PATH]}')
        with open(file[WRITE_PATH], 'wb') as out:
            log.info(f'Writing {file[NAME]} to {file[WRITE_PATH]}')
            out.write(file[DATA])
    except Exception as e:
        log_current_exception(traceback=False)
        raise Exception(f'Could not save {file[NAME]}')


def delete_file(file, data_dir):
    if READ_PATH in file:
        if file[READ_PATH].startswith(data_dir):
            log.warning(f'Not deleting {file[NAME]} as located within Data.Dir ({file[READ_PATH]})')
        else:
            slow_warning(f'Deleting {file[NAME]} from {file[READ_PATH]}')
            unlink(file[READ_PATH])


def upload_files(record, db, base, files=tuple(), nfiles=None, items=tuple(), progress=None,
                 unsafe=False, delete=False, replace=False):
    try:
        local_progress = ProgressTree(len(files), parent=progress)
    except TypeError:
        local_progress = ProgressTree(nfiles, parent=progress)
    with record.record_exceptions():
        with db.session_context() as s:
            data_dir = base_system_path(base, version=PERMANENT)
            check_items(s, items)
            for file in files:
                with local_progress.increment_or_complete():
                    try:
                        read_file(file)
                        hash_file(file)
                        parse_fit_data(file, items=items)
                        build_path(data_dir, file)
                        if replace: remove_previous_activity(s, file)
                        check_file(s, file, unsafe)
                        write_file(file)
                        record.info(f'Uploaded {file[NAME]} to {file[WRITE_PATH]}')
                        if delete: delete_file(file, data_dir)
                        if replace: remove_previous_files(file)
                    except SkipFile as e:
                        record.warning(e)
            local_progress.complete()  # catch no files case


def upload_files_and_update(record, sys, db, base, files=tuple(), nfiles=None, force=False, items=tuple(),
                            fast=False, unsafe=False, delete=False, replace=False):
    # this expects files to be a list of maps from name to stream (or an iterator, if nfiles provided)
    if unsafe:
        record.warning('Unsafe option enabled')
    with db.session_context() as s:
        n = ActivityJournal.number_of_activities(s)
        weight = 1 if force else max(1, int(sqrt(n)))
        log.debug(f'Weight statistics as {weight} ({n} entries)')
    progress = ProgressTree(1) if fast else SystemProgressTree(sys, UPLOAD, [1] * 5 + [weight])
    upload_files(record, db, base, files=files, nfiles=nfiles, items=items, progress=progress,
                 unsafe=unsafe, delete=delete, replace=replace)
    # todo - add record to pipelines?
    if not fast:
        run_activity_pipelines(sys, db, base, force=force, progress=progress)
        # run before and after so we know what exists before we update, and import what we read
        run_monitor_pipelines(sys, db, base, force=force, progress=progress)
        with db.session_context() as s:
            try:
                run_garmin(sys, s, progress=progress)
            except Exception as e:
                log.warning(f'Could not get data from Garmin: {e}')
        run_monitor_pipelines(sys, db, base, force=force, progress=progress)
        run_statistic_pipelines(sys, db, base, force=force, progress=progress)



import datetime as dt
from logging import getLogger
from re import compile

from sqlalchemy import desc, asc

from . import MultiProcCalculator, ActivityJournalCalculatorMixin
from .activity import ActivityCalculator
from ..names import ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED, MAX_MEAN_PE_M_ANY, FITNESS_D_ANY, _delta, TOTAL_CLIMB, \
    CLIMB_DISTANCE, CLIMB_ELEVATION, MAX_MED_HR_M_ANY, MIN_KM_TIME_ANY
from ...lib import local_time_to_time
from ...lib.log import log_current_exception
from ...sql import ActivityJournal, Timestamp, StatisticName, StatisticJournal, Achievement, ActivityGroup
from ...sql.tables.statistic import STATISTIC_JOURNAL_CLASSES, StatisticJournalFloat
from ...sql.types import short_cls
from ...sql.utils import add

log = getLogger(__name__)


class AchievementCalculator(ActivityJournalCalculatorMixin, MultiProcCalculator):

    def _delete(self, s):
        start, finish = self._start_finish(type=local_time_to_time)
        s.commit()   # so that we don't have any risk of having something in the session that can be deleted
        journal_ids = s.query(ActivityJournal.id)
        if start is not None:
            journal_ids = journal_ids.filter(ActivityJournal.finish > start)
        if finish is not None:
            journal_ids = journal_ids.filter(ActivityJournal.start < finish)
        q = s.query(Achievement).filter(Achievement.activity_journal_id.in_(journal_ids))
        n = q.count()
        log.debug(f'Deleting {n} achievements')
        q.delete(synchronize_session=False)
        Timestamp.clear_keys(s, journal_ids, owner=self.owner_out)

    def _startup(self, s):
        self._table = self._build_table(s)

    def _build_table(self, s):
        table = []
        self._append_like(table, s, 'fastest', 15, MIN_KM_TIME_ANY, ActivityCalculator)
        self._append_like(table, s, 'longest', 10, ACTIVE_DISTANCE, ActivityCalculator)
        self._append_like(table, s, 'longest', 10, ACTIVE_TIME, ActivityCalculator)
        self._append_like(table, s, 'fastest', 10, ACTIVE_SPEED, ActivityCalculator)
        self._append_like(table, s, 'highest', 3, MAX_MEAN_PE_M_ANY, ActivityCalculator)
        self._append_like(table, s, 'highest', 10, _delta(FITNESS_D_ANY), ActivityCalculator)
        self._append_like(table, s, 'highest', 10, TOTAL_CLIMB, ActivityCalculator)
        self._append_like(table, s, 'highest', 5, CLIMB_ELEVATION, ActivityCalculator)
        self._append_like(table, s, 'highest', 5, CLIMB_DISTANCE, ActivityCalculator)
        self._append_like(table, s, 'highest', 3, MAX_MED_HR_M_ANY, ActivityCalculator)
        return table

    def _append_like(self, table, s, superlative, score, pattern, owner):
        for statistic_name in s.query(StatisticName). \
                filter(StatisticName.name.like(pattern),
                       StatisticName.owner == owner):
            table.append((superlative, statistic_name, score))

    def _run_one(self, s, time_or_date):
        activity_journal = s.query(ActivityJournal).filter(ActivityJournal.start == time_or_date).one()
        with Timestamp(owner=self.owner_out, source=activity_journal).on_success(s):
            try:
                self._calculate_stats(s, activity_journal)
            except Exception as e:
                log.warning(f'No statistics on {time_or_date}: {e}')
                log_current_exception()

    def _calculate_stats(self, s, activity_journal):
        log.debug(f'Calculate for {activity_journal} on {activity_journal.start}')
        self._check_all(s, activity_journal)

    def _check_all(self, s, activity_journal):
        for (superlative, statistic_name, statistic_score) in self._table:
            # going down to a week gets way too many
            for (days, period, period_score) in [(50 * 365, 'of all time', 10),
                                                 (365, 'in a year', 5),
                                                 (30, 'in 30 days', 1)]:
                rank, achievement, all = self._check(s, activity_journal, superlative, statistic_name, days, period)
                if achievement:
                    score = statistic_score + period_score - rank + (4 if all else 0)
                    log.debug(f'{achievement}/{score}')
                    add(s, Achievement(text=achievement, sort=score, activity_journal=activity_journal))
                    break  # if month, also week

    def _build_query(self, s, activity_journal, statistic_name, days=0):
        less_is_better = 'min' in statistic_name.summaries
        order = asc if less_is_better else desc
        journal_class = STATISTIC_JOURNAL_CLASSES[statistic_name.statistic_journal_type]
        q = s.query(journal_class.value). \
            filter(StatisticJournal.statistic_name == statistic_name,
                   StatisticJournal.time >= activity_journal.start - dt.timedelta(days=days),
                   StatisticJournal.time < activity_journal.finish)
        if journal_class == StatisticJournalFloat:
            q = q.filter(journal_class.value != 0.0)
        q = q.order_by(order(journal_class.value))
        return q

    def _check(self, s, activity_journal, superlative, statistic_name, days, period):
        try:
            group = self.parse_group(statistic_name)
            # 4 so we know something worse
            best_values = self._build_query(s, activity_journal, statistic_name, days).limit(4).all()
            best_values = [x[0] for x in best_values]
            current_value = self._build_query(s, activity_journal, statistic_name).limit(1).scalar()
            for rank, adjective in enumerate(('%s', '2nd %s', '3rd %s')):
                description = adjective % superlative
                # +1 below so we don't give prizes for last
                if len(best_values) > rank+1 and current_value == best_values[rank]:
                    achievement = f'{description} {lower(statistic_name.name)} {period} (for {group})'
                    achievement = achievement[0].upper() + achievement[1:]
                    return rank, achievement, group == 'all'
        except Exception as e:
            log.warning(f'No achievement for {statistic_name}: {e}')
            log_current_exception()
        return 0, None, False

    @staticmethod
    def parse_group(statistic_name):
        # activity group is encoded into the statistic name constraint
        constraint = statistic_name.constraint
        if constraint.startswith(short_cls(ActivityGroup)):
            return lower(constraint.split('"')[1])
        else:
            return 'all'


def lower(text):
    def l(word):
        if compile(r'^[A-Z][a-z]+$').match(word):
            word = word[0].lower() + word[1:]
        return word
    return ' '.join(l(word) for word in text.split(' '))



import datetime as dt
import time as t
from logging import getLogger

from ..json import JsonResponse
from ...diary.database import read_date, read_schedule
from ...diary.views.web import rewrite_db
from ...lib import time_to_local_time
from ...lib.schedule import Schedule
from ...sql import ActivityJournal, StatisticJournal
from ...stats.display.activity import active_days, active_months

log = getLogger(__name__)


class Diary:

    FMT = ('%Y', '%Y-%m', '%Y-%m-%d')

    @staticmethod
    def read_diary(request, s, date):
        schedule, date = parse_date(date)
        if schedule == 'd':
            data = read_date(s, date)
        else:
            data = read_schedule(s, Schedule(schedule), date)
        return rewrite_db(list(data))

    def read_neighbour_activities(self, request, s, date):
        # used in the sidebar menu to advance/retreat to the next activity
        ymd = date.count('-')
        before = ActivityJournal.before_local_time(s, date)
        after = ActivityJournal.after_local_time(s, date)
        result = {}
        if before: result['before'] = time_to_local_time(before.start, self.FMT[ymd])
        if after: result['after'] = time_to_local_time(after.start, self.FMT[ymd])
        return JsonResponse(result)

    @staticmethod
    def read_active_days(request, s, month):
        return JsonResponse(active_days(s, month))

    @staticmethod
    def read_active_months(request, s, year):
        return JsonResponse(active_months(s, year))

    @staticmethod
    def write_statistics(request, s):
        # used to write modified fields back to the database
        data = request.json
        log.info(data)
        n = 0
        for key, value in data.items():
            try:
                id = int(key)
                journal = s.query(StatisticJournal).filter(StatisticJournal.id == id).one()
                journal.set(value)
                n += 1
            except Exception as e:
                log.error(f'Could not save {key}:{value}: {e}')
        s.commit()
        log.info(f'Saved {n} values')


def parse_date(date):
    for schedule, format in (('y', '%Y'), ('m', '%Y-%m'), ('d', '%Y-%m-%d')):
        try:
            return schedule, dt.date(*t.strptime(date, format)[:3])
        except:
            pass
    raise Exception(f'Cannot parse {date}')




from logging import getLogger

from ..stats.load import StatisticJournalLoader

log = getLogger(__name__)


def unlock(args, system, db):
    '''
## unlock

    > ch2 unlock

Remove the "dummy" entry from the database that is used to coordinate locking across processes.

This should not be needed in normal use.  DO NOT use when worker processes are still running.
    '''
    with db.session_context() as s:
        log.info('Removing dummy entry...')
        StatisticJournalLoader.unlock(s)
        log.info('Removed (if present)')




from logging import getLogger

import psutil as ps

from ...commands.args import mm, WEB, UPLOAD, TUI, LOG, BASE
from ...commands.upload import upload_files_and_update, STREAM, NAME
from ...lib.log import Record
from ...lib.workers import command_root

log = getLogger(__name__)


class Upload:

    def __init__(self, sys, db, base):
        self._sys = sys
        self._db = db
        self._base = base

    def __call__(self, request, s):
        files = [{NAME: file.filename, STREAM: file.stream} for file in request.files.getlist('files')]
        items = request.form.getlist('kit')
        # we do this in two stages
        # first, immediate saving of files while web browser waiting for response
        upload_files_and_update(Record(log), self._sys, self._db, self._base, files=files, items=items, fast=True)
        # second, start rest of ingest process in background
        # tui to avoid stdout appearing on web service output
        cmd = f'{command_root()} -v5 {mm(TUI)} {mm(BASE)} {self._base} {mm(LOG)} {WEB}-{UPLOAD}.log {UPLOAD}'
        log.info(f'Starting {cmd}')
        ps.Popen(args=cmd, shell=True)
        # wait so that the progress has time to kick in
        self._sys.wait_for_progress(UPLOAD)




from logging import getLogger
from subprocess import run
from tempfile import TemporaryDirectory
from tests import LogTestCase

from sqlalchemy.sql.functions import count

from ch2.commands.args import m, V, bootstrap_dir
from ch2.config.profile.acooke import acooke
from ch2.lib.date import to_date
from ch2.sql.tables.source import Source, Interval
from ch2.sql.tables.statistic import StatisticJournalText, StatisticJournal, StatisticJournalFloat, StatisticName, \
    StatisticJournalInteger, StatisticJournalType
from ch2.sql.tables.topic import DiaryTopicJournal, DiaryTopic
from ch2.sql.utils import add
from ch2.stats.calculate.summary import SummaryCalculator

log = getLogger(__name__)


# the idea here is to test the new database schema with sources etc
# so we configure a database then load some data, calculate some stats,
# and see if everything works as expected.

class TestSources(LogTestCase):

    def test_sources(self):

        with TemporaryDirectory() as f:

            args, sys, db = bootstrap_dir(f, m(V), '5', configurator=acooke)

            with db.session_context() as s:

                # add a diary entry

                journal = add(s, DiaryTopicJournal(date='2018-09-29'))
                cache = journal.cache(s)
                diary = s.query(DiaryTopic).filter(DiaryTopic.name == 'Status').one()
                fields = diary.fields
                self.assertEqual(len(fields), 6, list(enumerate(map(str, fields))))
                self.assertEqual(fields[0].statistic_name.name, 'Notes')
                self.assertEqual(fields[1].statistic_name.name, 'Weight', str(fields[1]))
                statistics = [cache[field] for field in fields]
                for statistic in statistics:
                    self.assertTrue(statistic.value is None, statistics)
                statistics[0].value = 'hello world'
                statistics[1].value = 64.5

            with db.session_context() as s:

                # check the diary entry was persisted

                journal = DiaryTopicJournal.get_or_add(s, '2018-09-29')
                cache = journal.cache(s)
                diary = s.query(DiaryTopic).filter(DiaryTopic.name == 'Status').one()
                fields = diary.fields
                self.assertEqual(len(fields), 6, list(enumerate(map(str, fields))))
                self.assertEqual(fields[0].statistic_name.name, 'Notes')
                self.assertEqual(fields[1].statistic_name.name, 'Weight', str(fields[1]))
                statistics = [cache[field] for field in fields]
                self.assertEqual(statistics[1].value, 64.5)
                self.assertEqual(statistics[1].type, StatisticJournalType.FLOAT)

            # generate summary stats

            SummaryCalculator(sys, db, schedule='m').run()
            SummaryCalculator(sys, db, schedule='y').run()

            with db.session_context() as s:

                # check the summary stats

                diary = s.query(DiaryTopic).filter(DiaryTopic.name == 'Status').one()
                weights = s.query(StatisticJournal).join(StatisticName). \
                               filter(StatisticName.owner == diary, StatisticName.name == 'Weight'). \
                               order_by(StatisticJournal.time).all()
                self.assertEqual(len(weights), 2)
                self.assertEqual(weights[1].value, 64.5)
                self.assertEqual(len(weights[1].measures), 2, weights[1].measures)
                self.assertEqual(weights[1].measures[0].rank, 1)
                self.assertEqual(weights[1].measures[0].percentile, 100, weights[1].measures[0].percentile)
                n = s.query(count(StatisticJournalFloat.id)).scalar()
                self.assertEqual(n, 6, n)
                n = s.query(count(StatisticJournalInteger.id)).scalar()
                self.assertEqual(n, 10, n)
                m_avg = s.query(StatisticJournalFloat).join(StatisticName). \
                    filter(StatisticName.name == 'Avg/Month Weight').one()
                self.assertEqual(m_avg.value, 64.5)
                y_avg = s.query(StatisticJournalFloat).join(StatisticName). \
                    filter(StatisticName.name == 'Avg/Year Weight').one()
                self.assertEqual(y_avg.value, 64.5)
                month = s.query(Interval).filter(Interval.schedule == 'm').one()
                self.assertEqual(month.start, to_date('2018-09-01'), month.start)
                self.assertEqual(month.finish, to_date('2018-10-01'), month.finish)

            with db.session_context() as s:

                # delete the diary entry

                journal = DiaryTopicJournal.get_or_add(s, '2018-09-29')
                s.delete(journal)

            with db.session_context() as s:

                # check the delete cascade

                self.assertEqual(s.query(count(DiaryTopicJournal.id)).scalar(), 1)
                # this should be zero because the Intervals were automatically deleted
                # (well, now +1 because there's an original default weight)
                for source in s.query(Source).all():
                    print(source)
                for journal in s.query(StatisticJournal).all():
                    print(journal)
                self.assertEqual(s.query(count(Source.id)).scalar(), 31, list(map(str, s.query(Source).all())))  # constants
                self.assertEqual(s.query(count(StatisticJournalText.id)).scalar(), 19, s.query(count(StatisticJournalText.id)).scalar())
                self.assertEqual(s.query(count(StatisticJournal.id)).scalar(), 32, s.query(count(StatisticJournal.id)).scalar())




import datetime as dt
from tests import LogTestCase

from ch2.lib.date import to_date
from ch2.lib.schedule import Schedule, DateOrdinals, POS_INFINITY


class TestRepeating(LogTestCase):

    def assert_str(self, x1, s):
        s1 = str(x1)
        self.assertEqual(s1, s)
        x2 = Schedule(s1)
        s2 = str(x2)
        self.assertEqual(s1, s2)

    def assert_bad(self, f, *args):
        with self.assertRaises(Exception):
            f(*args)

    def test_specification(self):
        # from the spec doc
        self.assert_str(Schedule('m[2Tue]'), 'm[2tue]')
        self.assert_str(Schedule('2m[2Tue]'), '2m[2tue]')
        self.assert_str(Schedule('1/2m[2Tue]'), '1/2m[2tue]')
        self.assert_str(Schedule('m[2Tue]1970-01-01-1970-12-31'), 'm[2tue]1970-01-01-1970-12-31')
        self.assert_str(Schedule('2018-07-05/5d[]'), '2/5d')
        self.assert_str(Schedule('d2018-07-05'), '2018-07-05')
        # others
        self.assert_str(Schedule('1/2w[1]2018-01-01-'), '1/2w[1]2018-01-01-')
        self.assert_str(Schedule('1/2w[2Mon,Fri,2,3]-1970-01-01'), '1/2w[2,3,fri,2mon]-1970-01-01')
        # some errors
        self.assert_bad(Schedule, '1/2[]2018-01-01-')   # must specify type
        self.assert_bad(Schedule, '1/2w[1d]2018-01-01-')  # no longer support type in location
        # some additional shortcuts for easier config
        self.assert_str(Schedule('-2019-10-07'), '-2019-10-07')
        self.assert_str(Schedule('2019-10-07-'), '2019-10-07-')
        self.assert_str(Schedule('2019-10-07'), '2019-10-07')

    def assert_at(self, spec, date, at_location):
        date = to_date(date)
        self.assertEqual(Schedule(spec).at_location(date), at_location)

    def test_day(self):
        self.assert_at('d', '2018-07-06', True)
        self.assert_at('d', '2018-07-07', True)
        self.assert_at('2d', '2018-07-06', True)
        self.assert_at('2d', '2018-07-07', True)
        self.assert_at('2d[1]', '2018-07-06', True)
        self.assert_at('2d[2]', '2018-07-06', False)
        self.assert_at('2d[3]', '2018-07-06', False)
        self.assert_at('2d[1]', '2018-07-07', False)
        self.assert_at('2d[2]', '2018-07-07', True)
        self.assert_at('2d[Fri]', '2018-07-06', True)
        self.assert_at('2d[Mon]', '2018-07-06', False)
        self.assert_at('2d[Fri]2018-07-06', '2018-07-06', True)
        self.assert_at('2d[Fri]2018-07-07', '2018-07-06',  False)
        self.assert_at('2d[Fri]2018-07-06-2018-07-07', '2018-07-06', True)
        self.assert_at('2d[Fri]2018-07-07-2018-07-08', '2018-07-06', False)
        self.assert_at('2d[Fri]2018-07-05-2018-07-06', '2018-07-06', False)
        self.assert_at('2018-10-10/3d[1,fri]', '2018-10-10', True)
        self.assert_at('2018-10-10/3d[1,fri]', '2018-10-11', False)
        self.assert_at('2018-10-10/3d[1,fri]', '2018-10-12', True)

    def test_week(self):
        self.assert_at('2018-07-06/w[Fri]', '2018-07-06', True)
        self.assert_at('2018-07-06/w[Fri]', '2018-07-05', False)
        self.assert_at('2018-07-06/w[Fri]', '2018-07-07', False)
        self.assert_at('2018-07-06/w[1]', '2018-07-02', True)
        self.assert_at('2018-07-06/w[1]', '2018-07-01', False)
        self.assert_at('2018-07-06/w[1]', '2018-07-03', False)
        self.assert_at('2018-07-06/w', '2018-07-02', True)
        self.assert_at('2018-07-06/w', '2018-07-01', True)
        self.assert_at('2018-07-06/w', '2018-07-03', True)
        # bug in diary
        self.assert_at('0/1w[1sun]', '2018-07-29', True)

    def test_month(self):
        self.assert_at('m[Sat]', '2018-07-07', True)
        self.assert_at('m[2]', '2018-07-02', True)
        self.assert_at('m[2]', '2018-07-01', False)
        self.assert_at('m', '2018-07-02', True)
        # numbering is different for months.  "2nd sunday" may not be in 2nd week.
        self.assert_at('m[1sat]', '2018-09-01', True)
        self.assert_at('m[1sun]', '2018-09-02', True)
        self.assert_at('m[1mon]', '2018-09-03', True)
        self.assert_at('m[1fri]', '2018-09-07', True)
        self.assert_at('m[2sat]', '2018-09-08', True)
        # bug seen in output from test_schedule command
        self.assert_at('m[mon,tue,5]', '2018-10-08', True)

    def test_year(self):
        self.assert_at('2018-07-07/y', '2018-07-02', True)

    def test_start_finish(self):
        p = Schedule('2018-07-08/2d2018-07-08-2018-07-09')
        self.assertEqual(p.start, dt.date(2018, 7, 8))
        p.start = None
        self.assertTrue(p.start is None)
        self.assertEqual(p.finish, dt.date(2018, 7, 9))

    def test_ordinals(self):
        d = dt.date(2018, 7, 25)
        o = DateOrdinals(d)
        self.assertEqual(o.dow, 2)  # wed

    def test_frame_start(self):
        s = Schedule('2018-01-01/2y')
        self.assertEqual(s.start_of_frame('2018-01-02'), to_date('2018-01-01'))
        self.assertEqual(s.start_of_frame('2017-01-02'), to_date('2016-01-01'))

    def test_extended(self):
        # from the spec doc
        self.assert_str(Schedule('x2018-10-07-'), 'x2018-10-07-')
        self.assert_str(Schedule('x[1,2,3]2020-03-06-'), 'x[1,2,3]2020-03-06-')
        self.assert_str(Schedule('x[fri]2020-03-06-'), 'x[fri]2020-03-06-')
        self.assert_bad(Schedule, 'x[1,2,3]')  # no range start
        self.assert_bad(Schedule, '2018-10-07x')  # no offset
        self.assert_bad(Schedule, '2x')  # no multiple repeat
        self.assert_at('x', '2018-07-06', True)
        # single day range
        self.assert_at('x2018-07-06', '2018-07-05', False)
        self.assert_at('x2018-07-06', '2018-07-06', True)
        self.assert_at('x2018-07-06', '2018-07-07', False)
        # from date onwards
        self.assert_at('x2018-07-06-', '2018-07-05', False)
        self.assert_at('x2018-07-06-', '2018-07-06', True)
        self.assert_at('x2018-07-06-', '2018-07-07', True)
        # until date
        self.assert_at('x-2018-07-06', '2018-07-05', True)
        self.assert_at('x-2018-07-06', '2018-07-06', False)
        self.assert_at('x-2018-07-06', '2018-07-07', False)
        # numbered locations with start of range
        self.assert_at('x[1,2,3]2018-07-06-', '2018-07-05', False)
        self.assert_at('x[1,2,3]2018-07-06-', '2018-07-06', True)
        self.assert_at('x[1,2,3]2018-07-06-', '2018-07-07', True)
        self.assert_at('x[1,2,3]2018-07-06-', '2018-07-08', True)
        self.assert_at('x[1,2,3]2018-07-06-', '2018-07-09', False)
        # named locations with start of range
        self.assert_at('x[fri]2018-07-06-', '2018-07-05', False)
        self.assert_at('x[fri]2018-07-06-', '2018-07-06', True)
        self.assert_at('x[fri]2018-07-06-', '2018-07-07', False)
        # named locations with no range
        self.assert_at('x[fri]', '2018-07-05', False)
        self.assert_at('x[fri]', '2018-07-06', True)
        self.assert_at('x[fri]', '2018-07-07', False)
        s = Schedule('x')
        self.assertEqual(s.start_of_frame('2018-01-02'), to_date('1970-01-02'))
        self.assertEqual(s.start, None)
        self.assertEqual(s.finish, None)
        # this is not None because it is used to set end dates for repeating intervals
        self.assertEqual(s.next_frame('2018-01-02'), POS_INFINITY)

    def test_error_messages(self):
        with self.assertRaisesRegex(Exception, r'Cannot parse "z"'):
            Schedule('z')
        with self.assertRaisesRegex(Exception, r'Cannot parse "z2018-10-07-"'):
            Schedule('z2018-10-07-')
        with self.assertRaisesRegex(Exception, r'Numbered locations in daily frames not supported'):
            Schedule('d[1mon]')
        with self.assertRaisesRegex(Exception, r'Repetition and offset not supported by extended frames'):
            Schedule('2x')




from collections import defaultdict
from itertools import zip_longest, groupby
from logging import getLogger
from os import getpid, close, execl, execle
from os.path import split, realpath, normpath, expanduser
from pprint import PrettyPrinter
from sys import executable, argv
from time import sleep

from psutil import Process

from ..stats.names import M, KM, PC, W


log = getLogger(__name__)


PALETTE_RAINBOW = [
    ('plain', 'light gray', 'black'), ('plain-focus', 'white', 'black'),
    ('em', 'white', 'black'),
    ('label', 'dark gray', 'black'),
    ('selected', 'black', 'light gray'), ('selected-focus', 'black', 'white'),
    ('unimportant', 'dark blue', 'black'), ('unimportant-focus', 'light blue', 'black'),
    ('error', 'dark red', 'black'), ('error-focus', 'light red', 'black'),
    ('bar', 'dark gray', 'black'), ('bar-focus', 'dark gray', 'black'),
    ('rank-1', 'light red', 'black'), ('rank-2', 'yellow', 'black'), ('rank-3', 'dark green', 'black'),
    ('rank-4', 'light blue', 'black'), ('rank-5', 'light gray', 'black'),
    ('zone-6', 'black', 'white'), ('zone-5', 'black', 'light red'), ('zone-4', 'black', 'yellow'),
    ('zone-3', 'black', 'dark green'), ('zone-2', 'black', 'light blue'), ('zone-1', 'black', 'light gray'),
    ('quintile-1', 'light gray', 'black'), ('quintile-2', 'light blue', 'black'), ('quintile-3', 'dark green', 'black'),
    ('quintile-4', 'yellow', 'black'), ('quintile-5', 'light red', 'black'),
]

PALETTE_MINIMAL = [
    ('plain', 'light gray', 'black'), ('plain-focus', 'white', 'black'),
    ('em', 'white', 'black'),
    ('label', 'dark gray', 'black'),
    ('selected', 'black', 'light gray'), ('selected-focus', 'black', 'white'),
    ('unimportant', 'dark blue', 'black'), ('unimportant-focus', 'light blue', 'black'),
    ('error', 'dark red', 'black'), ('error-focus', 'light red', 'black'),
    ('bar', 'dark gray', 'black'), ('bar-focus', 'dark gray', 'black'),
    ('rank-1', 'white', 'black'), ('rank-2', 'light gray', 'black'), ('rank-3', 'light gray', 'black'),
    ('rank-4', 'light gray', 'black'), ('rank-5', 'light gray', 'black'),
    ('zone-6', 'black', 'white'), ('zone-5', 'black', 'white'), ('zone-4', 'black', 'white'),
    ('zone-3', 'black', 'light gray'), ('zone-2', 'black', 'dark gray'), ('zone-1', 'black', 'dark gray'),
    ('quintile-1', 'light gray', 'black'), ('quintile-2', 'light gray', 'black'), ('quintile-3', 'light gray', 'black'),
    ('quintile-4', 'light gray', 'black'), ('quintile-5', 'white', 'black'),
]

PALETTE = PALETTE_MINIMAL



def sign(x):
    if x == 0:
        return 0
    elif x > 0:
        return 1
    else:
        return -1


def sigfig(value, n=2):
    scale = 1
    while value >= 10 ** n:
        value /= 10
        scale *= 10
    while value < 10 ** (n-1):
        value *= 10
        scale /= 10
    return int(0.5 + value) * scale


def em(text):
    return 'em', text


def error(text):
    return 'error', text


def label(text):
    return 'label', text


def force_iterable(data):
    try:
        iter(data)
        return data
    except TypeError:
        return [data]


def unique(elements, key=lambda x: x):
    known = set()
    for element in elements:
        value = key(element)
        if value not in known:
            known.add(value)
            yield element


def interleave(sep, iter):
    for started, value in enumerate(iter):
        if started:
            yield sep
        yield value


# https://docs.python.org/3/library/itertools.html#itertools-recipes
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


PP = PrettyPrinter(indent=0, depth=1, width=80, compact=True)


def short_str(x):
    text = PP.pformat(x)
    lines = text.splitlines(False)
    if len(lines) > 1:
        return lines[0][:20] + '...' + lines[-1][-20:]
    else:
        return text


def format_metres(dist):
    if dist < 1000:
        return str(int(dist)) + M
    else:
        return format_km(dist / 1000)


def format_km(dist):
    return f'{dist:.1f}{KM}'


def format_percent(pc):
    if pc > 0.5:
        return f'{pc:.1f}{PC}'
    else:
        return f'{pc:.2f}{PC}'


def format_watts(power):
    return str(int(power)) + W


def groupby_tuple(iterable, key=None):
    for name, group in groupby(iterable, key=key):
        yield name, tuple(group)


def group_to_dict(iterable):
    'Expects iterable to be (key, value) tuples.'
    d = defaultdict(list)
    for key, value in iterable:
        d[key].append(value)
    return d


def drop_trailing_slash(path):
    left, right = split(path)
    if not right:
        return left
    else:
        return path


def inside_interval(lo, value, hi):
    if lo is None:
        if hi is None:
            return True
        else:
            return value < hi
    else:
        if hi is None:
            return lo <= value
        else:
            return lo <= value < hi


def restart_self():
    # https://stackoverflow.com/questions/11329917/restart-python-script-from-within-itself
    log.info('Shutting down')
    try:
        p = Process(getpid())
        for handler in p.open_files() + p.connections():
            close(handler.fd)
    except Exception as e:
        log.warning(e)
    python = executable
    args = argv
    # weird hack that i don't understand
    args = ['-m', 'ch2'] + args[1:]
    log.info(f'Restarting {python} {args}')
    execl(python, python, *args)
    # no need to exit as we do not return


def clean_path(path):
    return realpath(normpath(expanduser(path)))


def slow_warning(msg, n=3, pause=1):
    for _ in range(3):
        log.warning(msg)
        sleep(pause)




from contextlib import contextmanager
from logging import getLogger, DEBUG, Formatter, INFO, StreamHandler, WARNING
from logging.handlers import RotatingFileHandler
from sys import exc_info
from traceback import format_tb

from ..commands.args import COMMAND, LOGS, PROGNAME, VERBOSITY, LOG, TUI


log = getLogger(__name__)


def make_log_from_args(args):

    name = args[LOG] if LOG in args and args[LOG] else (
            (args[COMMAND] if COMMAND in args and args[COMMAND] else PROGNAME) + f'.{LOG}')
    path = args.system_path(LOGS, name)

    make_log(path, verbosity=args[VERBOSITY], tui=args[TUI])


def make_log(path, verbosity=4, tui=False):

    if not getLogger('ch2').handlers:

        level = 10 * (6 - verbosity)

        file_formatter = Formatter('%(levelname)-8s %(asctime)s: %(message)s')
        file_handler = RotatingFileHandler(path, maxBytes=1e6, backupCount=10)
        file_handler.setLevel(DEBUG)
        file_handler.setFormatter(file_formatter)

        slog = getLogger('sqlalchemy')
        slog.setLevel(WARNING)
        slog.addHandler(file_handler)

        mlog = getLogger('matplotlib')
        mlog.setLevel(INFO)
        mlog.addHandler(file_handler)

        blog = getLogger('bokeh')
        blog.setLevel(DEBUG)
        blog.addHandler(file_handler)

        tlog = getLogger('tornado')
        tlog.setLevel(INFO)
        tlog.addHandler(file_handler)

        sslog = getLogger('sentinelsat')
        sslog.setLevel(DEBUG)
        sslog.addHandler(file_handler)

        wlog = getLogger('werkzeug')
        wlog.setLevel(DEBUG)
        wlog.addHandler(file_handler)

        clog = getLogger('ch2')
        clog.setLevel(DEBUG)
        clog.addHandler(file_handler)

        # capture logging from an executing module, if one exists
        xlog = getLogger('__main__')
        xlog.setLevel(DEBUG)
        xlog.addHandler(file_handler)

        if not tui:
            stderr_formatter = Formatter('%(levelname)8s: %(message)s')
            stderr_handler = StreamHandler()
            stderr_handler.setLevel(level)
            stderr_handler.setFormatter(stderr_formatter)
            blog.addHandler(stderr_handler)
            tlog.addHandler(stderr_handler)
            wlog.addHandler(stderr_handler)
            clog.addHandler(stderr_handler)
            xlog.addHandler(stderr_handler)


def log_current_exception(traceback=True):
    t, e, tb = exc_info()
    try:
        log.debug(f'Exception: {e}')
    except:
        pass
    log.debug(f'Type: {t}')
    if traceback:
        log.debug('Traceback:\n' + ''.join(format_tb(tb)))


class Record:

    def __init__(self, log):
        self._log = log
        self._warning = []
        self._info = []

    def warning(self, msg):
        self._log.warning(msg)
        self._warning.append(msg)

    def info(self, msg):
        self._log.info(msg)
        self._info.append(msg)

    def raise_(self, msg):
        self.warning(msg)
        raise Exception(msg)

    @contextmanager
    def record_exceptions(self):
        try:
            yield
        except Exception as e:
            self.warning(e)
            raise

    def json(self):
        return {'warning': self._warning,
                'info': self._info}



import numpy as np
import pandas as pd

from .heart_rate import BC_ZONES
from ..lib.data import interpolate_freq
from ..lib.date import to_time
from ..stats.names import HEART_RATE, FTHR, HR_ZONE, HR_IMPULSE_10


def hr_zone(heart_rate_df, fthr_df, pc_fthr_zones=BC_ZONES, heart_rate=HEART_RATE, hr_zone=HR_ZONE):
    '''
    mutate input df to include hr zone

    this can be used for a huge slew of data, or for an individual activity.
    '''
    fthrs = sorted([(time, row[FTHR]) for time, row in fthr_df.dropna().iterrows()], reverse=True)
    if not fthrs:
        raise Exception(f'No {FTHR} defined')
    fthrs = fthrs + [(to_time('2100'), None)]
    fthrs = [(a[0], b[0], a[1]) for a, b in zip(fthrs, fthrs[1:])]
    heart_rate_df[hr_zone] = np.nan
    for start, finish, fthr in fthrs:  # start is inclusive
        start, finish = pd.to_datetime(start, utc=True), pd.to_datetime(finish, utc=True)
        zones = [x * fthr / 100.0 for x in pc_fthr_zones]
        for zone, upper in enumerate(zones + [None], start=1):
            if zone == 1:
                # we don't try to distinguish zones below 1
                heart_rate_df.loc[(heart_rate_df.index >= start) & (heart_rate_df.index < finish) &
                                  (heart_rate_df[heart_rate] <= upper),
                                  [hr_zone]] = zone
            elif not upper:
                # or above 5
                heart_rate_df.loc[(heart_rate_df.index >= start) & (heart_rate_df.index < finish) &
                                  (heart_rate_df[heart_rate] > lower),
                                  [hr_zone]] = zone
            else:
                hrz = ((heart_rate_df.loc[(heart_rate_df.index >= start) & (heart_rate_df.index < finish) &
                                          (heart_rate_df[heart_rate] > lower) & (heart_rate_df[heart_rate] <= upper),
                                          [heart_rate]] - lower) / (upper - lower)) + zone
                # .values below from
                # https://stackoverflow.com/questions/12307099/modifying-a-subset-of-rows-in-a-pandas-dataframe
                # i do not understand why it is needed...
                heart_rate_df.loc[(heart_rate_df.index >= start) & (heart_rate_df.index < finish) &
                                  (heart_rate_df[heart_rate] > lower) & (heart_rate_df[heart_rate] <= upper),
                                  [hr_zone]] = hrz.values
            lower = upper


def impulse_10(hr_zone_df, impulse, hr_zone=HR_ZONE):
    '''
    interpolate HR to 10s values then calculate impulse using model parameters.

    this can be used for a huge slew of data, or for an individual activity.
    '''
    if hr_zone_df.empty:
        impulse_df = pd.DataFrame(columns=[HR_IMPULSE_10])
    else:
        impulse_df = interpolate_freq(hr_zone_df.loc[:, [hr_zone]], '10s',
                                      method='index', limit=int(0.5 + impulse.max_secs / 10)).dropna()
        impulse_df[HR_IMPULSE_10] = (impulse_df[hr_zone] - impulse.zero) / (impulse.one - impulse.zero)
        impulse_df[HR_IMPULSE_10].clip(lower=0, inplace=True)
        impulse_df[HR_IMPULSE_10] = impulse_df[HR_IMPULSE_10] ** impulse.gamma
        impulse_df.drop(columns=[hr_zone], inplace=True)
    return impulse_df


if __name__ == '__main__':
    from ..sql.database import connect
    from ..stats.calculate.response import HRImpulse
    from ..data import statistics
    _, db = connect(['-v5'])
    with db.session_context() as s:
        impulse = HRImpulse(dest_name='Test Impulse', gamma=1.0, zero=2, one=6, max_secs=60)
        hr_df = statistics(s, HR_ZONE)
        print(hr_df.describe())
        print(hr_df)
        impulse_df = impulse_10(hr_df, impulse)
        print(impulse_df.describe())
        print(impulse_df)




from bokeh.models import PanTool, ZoomInTool, ZoomOutTool, ResetTool, HoverTool, LinearAxis
from bokeh.plotting import figure

from .line import multi_dot_plot, dot_plotter, comb_plotter, DEFAULT_BACKEND
from .utils import make_range, evenly_spaced_hues, tooltip
from ..frame import related_statistics
from ...stats.names import ACTIVE_TIME, ACTIVE_DISTANCE, TIME, ACTIVE_TIME_H, ACTIVE_DISTANCE_KM, LOCAL_TIME, _slash, \
    H, KM, ACTIVITY_GROUP, like, _delta, FITNESS_D_ANY, FATIGUE_D_ANY


def std_distance_time_plot(nx, ny, source, x_range=None, output_backend=DEFAULT_BACKEND):
    groups = [group for statistic, group in related_statistics(source, ACTIVE_TIME)]
    if not groups:
        # original monochrome plot
        return multi_dot_plot(nx, ny, TIME, [ACTIVE_TIME_H, ACTIVE_DISTANCE_KM], source,
                              ['black', 'grey'], alphas=[1, 0.5], x_range=x_range, rescale=True)
    times = [f'{ACTIVE_TIME} ({group})' for group in groups]
    distances = [f'{ACTIVE_DISTANCE} ({group})' for group in groups]
    time_y_range = make_range(source[ACTIVE_TIME_H])
    distance_y_range = make_range(source[ACTIVE_DISTANCE_KM])
    colours = list(evenly_spaced_hues(len(groups)))
    tooltip_names = [ACTIVE_TIME_H, ACTIVE_DISTANCE_KM, ACTIVITY_GROUP, LOCAL_TIME]
    tooltip_names += [name for name in like(_delta(FITNESS_D_ANY), source.columns) if '(' not in name]
    tooltip_names += [name for name in like(_delta(FATIGUE_D_ANY), source.columns) if '(' not in name]
    tools = [PanTool(dimensions='width'),
             ZoomInTool(dimensions='width'), ZoomOutTool(dimensions='width'),
             ResetTool(),
             HoverTool(tooltips=[tooltip(name) for name in tooltip_names], names=['with_hover'])]
    f = figure(output_backend=output_backend, plot_width=nx, plot_height=ny, x_axis_type='datetime', tools=tools)
    f.yaxis.axis_label = f'lines - {ACTIVE_TIME_H}'
    f.y_range = time_y_range
    f.extra_y_ranges = {ACTIVE_DISTANCE: distance_y_range}
    f.add_layout(LinearAxis(y_range_name=ACTIVE_DISTANCE, axis_label=f'dots - {ACTIVE_DISTANCE_KM}'), 'right')
    plotter = comb_plotter()
    for time, colour in zip(times, colours):
        time_h = _slash(time, H)
        source[time_h] = source[time] / 3600
        plotter(f, x=TIME, y=time_h, source=source, color=colour, alpha=1)
    plotter = dot_plotter()
    for distance, colour in zip(distances, colours):
        distance_km = _slash(distance, KM)
        source[distance_km] = source[distance] / 1000
        plotter(f, x=TIME, y=distance_km, source=source, color=colour, alpha=1, name='with_hover',
                y_range_name=ACTIVE_DISTANCE)
    f.xaxis.axis_label = TIME
    f.toolbar.logo = None
    if ny < 300: f.toolbar_location = None
    if x_range: f.x_range = x_range
    return f




from .achievement import Achievement
from .activity import ActivityGroup, ActivityTimespan, ActivityJournal, ActivityBookmark
from .constant import Constant
from .file import FileScan, FileHash
from .kit import KitGroup, KitItem, KitComponent, KitModel
from .monitor import MonitorJournal
from .nearby import ActivitySimilarity, ActivityNearby
from .pipeline import Pipeline, PipelineType
from .segment import Segment, SegmentJournal
from .source import Source, Interval, NoStatistics, Dummy, Composite, CompositeComponent
from .statistic import StatisticName, StatisticJournalFloat, StatisticJournalText, StatisticJournalInteger, \
    StatisticJournalTimestamp, StatisticJournal, StatisticMeasure, StatisticJournalType
from .system import SystemConstant, Process
from .timestamp import Timestamp
from .topic import DiaryTopicJournal, DiaryTopic, DiaryTopicField, ActivityTopicJournal, ActivityTopic, \
    ActivityTopicField




from logging import getLogger

from sentinelsat import SentinelAPI
from shapely.geometry import MultiPoint, box

from ..data import activity_statistics
from ..stats.names import LATITUDE, LONGITUDE

log = getLogger(__name__)


def query_activity(s, user, passwd, local_time, activity_group_name, margin=0.1):
    df = activity_statistics(s, LATITUDE, LONGITUDE, local_time=local_time, activity_group_name=activity_group_name)
    df = df.dropna()
    footprint = MultiPoint(df.apply(lambda row: (row[LONGITUDE], row[LATITUDE]), axis='columns')).convex_hull
    minx, miny, maxx, maxy = footprint.bounds
    dx, dy = maxx - minx, maxy - miny
    bbox = box(minx - margin * dx, miny - margin * dy, maxx + margin * dx, maxy + margin * dy)
    api = SentinelAPI(user, passwd, 'https://scihub.copernicus.eu/dhus')
    products = api.query(area=bbox, platformname='Sentinel-2', date=('NOW-1YEAR', 'NOW'),
                         processinglevel='Level-2A', cloudcoverpercentage=(0, 10))
    return api, products, bbox, df




from collections import defaultdict
from glob import glob
from logging import getLogger
from os.path import join, exists

import rasterio as rio
from rasterio.enums import Resampling
from rasterio.features import shapes
from rasterio.mask import mask
from rasterio.merge import merge
from rasterio.warp import calculate_default_transform, reproject, transform_geom
from shapely.geometry import Polygon, mapping, box

from ..lib.image import create_image, RGB, GTIFF, DRIVER, PHOTOMETRIC, DTYPES, COUNT, TRANSFORM, WIDTH, HEIGHT, CRS
from ..lib import drop_trailing_slash, median

log = getLogger(__name__)


def read_band(path, band):
    files = glob(join(path, '*.SAFE', 'GRANULE', '*', 'IMG_DATA', 'R10m', f'*_{band}_10m.jp2'))
    if len(files) != 1:
        raise Exception(f'Unexpected structure under {path} for band {band}')
    band = rio.open(files[0])
    log.debug(f'Opened {files[0]}')
    log.debug(band.profile)
    return band


def read_rgb(path):
    b = read_band(path, 'B02')
    g = read_band(path, 'B03')
    r = read_band(path, 'B04')
    return r, g, b


def create_rgb(path):
    '''
    The path points to a directory; we create an RGB composite from the data in the directory and save to the
    same path with '.tiff' appended.
    See https://towardsdatascience.com/satellite-imagery-access-and-analysis-in-python-jupyter-notebooks-387971ece84b
    '''
    tiff = drop_trailing_slash(path) + '.tiff'
    if exists(tiff):
        log.warning(f'{tiff} already exists')
    else:
        log.debug('Reading layers')
        r, g, b = read_rgb(path)
        profile = r.profile
        profile.update({DRIVER: GTIFF,
                        COUNT: 3,
                        DTYPES: (r.dtypes[0], g.dtypes[0], b.dtypes[0]),
                        PHOTOMETRIC: RGB})
        with rio.open(tiff, 'w', **profile) as out:
            log.debug('Writing red layer')
            out.write(r.read(1), 1)
            log.debug('Writing green layer')
            out.write(g.read(1), 2)
            log.debug('Writing blue layer')
            out.write(b.read(1), 3)
            out.close()
        log.debug(f'Wrote {path}')
    return tiff


def first(images):
    return images[0]


def split(images, choose=first):
    reference = choose(images)
    return reference, [image for image in images if image is not reference]


def force_same_crs(images, choose=first):
    reference, rest = split(images, choose)
    same = [reference]
    while rest:
        image = rest.pop()
        if image.crs != reference.crs:
            log.info(f'Converting {image.crs} to {reference.crs}')
            image = reproject_to_memory(image, reference.crs)
        same.append(image)
    return same


def reproject_to_memory(image, crs):
    # https://rasterio.readthedocs.io/en/stable/topics/reproject.html
    transform, width, height = calculate_default_transform(image.crs, crs, image.width, image.height, *image.bounds)
    profile = image.profile.copy()
    profile.update({CRS: crs,
                    TRANSFORM: transform,
                    WIDTH: width,
                    HEIGHT: height})
    transformed = rio.MemoryFile().open(**profile)
    for i in range(1, image.count + 1):
        reproject(source=rio.band(image, i),
                  destination=rio.band(transformed, i),
                  src_transform=image.transform,
                  src_crs=image.crs,
                  dst_transform=transform,
                  dst_crs=crs,
                  resampling=Resampling.nearest)
    return transformed


def shape_to_polygon(shape):  # to do - shapely shape() should do this?
    geometry, value = shape
    if geometry['type'] != 'Polygon':
        raise Exception(f'Unexpected geometry: {shape}')
    return Polygon(*geometry['coordinates'])


def most_intersecting(footprints, targets, candidates, choose=first):
    score_to_images = defaultdict(list)
    for candidate in candidates:
        score = sum(1 if footprints[target].intersection(footprints[candidate]) else 0 for target in targets)
        score_to_images[score].append(candidate)
    scores = sorted(score_to_images.keys(), reverse=True)
    if scores[0] == 0:
        raise Exception('Disjoint images')
    return choose(score_to_images[scores[0]])


def crop_to_shape(image, shape):
    cropped, transform = mask(image, [shape], crop=True)
    return create_image(image, cropped, transform)


def measure_scaling(footprints, targets, candidate):
    scalings = []
    for target in targets:
        overlap = footprints[target].intersection(footprints[candidate])
        if overlap:
            log.debug(f'{target} and {candidate} overlap at {overlap}')
            crop_target = crop_to_shape(target, overlap)
            crop_candidate = crop_to_shape(candidate, overlap)
            scalings.append(sorted(crop_target.read(i).mean() / crop_candidate.read(i).mean() for i in range(1, 4))[1])
    if not scalings:
        raise Exception('Disjoint images')
    return median(scalings)


def force_same_scaling(images, choose=first):
    footprints = {image: shape_to_polygon(next(shapes(image.dataset_mask(), transform=image.transform)))
                  for image in images}
    reference, rest = split(images, choose)
    scaled = [reference]
    while rest:
        candidate = most_intersecting(footprints, scaled, rest)
        rest.remove(candidate)
        scale = measure_scaling(footprints, scaled, candidate)
        log.info(f'Scaling {candidate} by {scale}')
        profile = candidate.profile
        profile.update({'photometric': 'RGB'})
        copy = rio.MemoryFile().open(**profile)
        types = {i: dtype for i, dtype in zip(candidate.indexes, candidate.dtypes)}
        for i in range(1, 4):
            copy.write((candidate.read(i) * scale).astype(types[i]), i)
        scaled.append(copy)
    return scaled


def combine_images(images, choose=first):
    log.debug(f'Combining {images}')
    images = force_same_crs(images, choose=choose)
    images = force_same_scaling(images, choose=choose)
    return create_image(images[0], *merge(images))


def crop_to_box(image, gps_bbox):
    '''
    This differs from crop to shape in that:
    1 - the bbox coords are WGS84 lat lon (GPS)
    2 - we want the result to be a 'square' image (not a diamond)
    (we're assuming the image is more-or-less aligned N/S)
    '''
    # for some reason the gps_bbox cannot be a shapely object; it has to be a geojson dict
    crs_bbox = transform_geom('WGS84', image.crs, mapping(gps_bbox))
    pixel_bbox = [image.index(*xy) for xy in crs_bbox['coordinates'][0]]
    xs, ys = zip(*pixel_bbox)
    # maybe we should do something smarter here to ensure we always have data?
    # like take the miniumum of the maximal values, etc.
    pixel_bbox = box(min(xs), min(ys), max(xs), max(ys))
    crs_bbox = Polygon([image.xy(*xy) for xy in pixel_bbox.exterior.coords])
    return crop_to_shape(image, crs_bbox)






import webbrowser
from logging import getLogger

from .args import SUB_COMMAND, SERVICE, STOP, START, STATUS
from ..web.server import WebController


log = getLogger(__name__)


def web(args, system, db):
    '''
## web

    > ch2 web start

Start the web server.

    > ch2 web status

Indicate whether the server is running or not.

    > ch2 web stop

Stop the server.
    '''
    cmd = args[SUB_COMMAND]
    controller = WebController(args, system, db)
    if cmd == SERVICE:
        controller.service()
    elif cmd == STATUS:
        controller.status()
    elif cmd == START:
        controller.start(restart=True)
        webbrowser.open(controller.connection_url(), autoraise=False)
        log.info(controller.connection_url())
    elif cmd == STOP:
        controller.stop()



from logging import getLogger

from . import copy_statistic_journal
from ...lib.log import log_current_exception
from ...sql import StatisticName, Constant, StatisticJournal

log = getLogger(__name__)


def upgrade_constant(record, old, new):
    if not constant_imported(record, new):
        with old.session_context() as old_s:
            copy_constants(record, old_s, old, new)
            list_missing_constants(record, new)


def constant_imported(record, new):
    with new.session_context() as new_s:
        # imported if NO undefined constants (more lenient than other imports)
        return not bool(new_s.query(Constant).
                        join(StatisticName).
                        outerjoin(StatisticJournal).
                        filter(StatisticJournal.id == None).count())


def list_missing_constants(record, new):
    with new.session_context() as new_s:
        for new_constant in new_s.query(Constant). \
                join(StatisticName). \
                outerjoin(StatisticJournal). \
                filter(StatisticJournal.id == None).all():
            record.warning(f'Constant {new_constant.name} still missing value(s)')


def copy_constants(record, old_s, old, new):
    constant = old.meta.tables['constant']
    for old_constant in old_s.query(constant).all():
        try:
            statistic_name = old.meta.tables['statistic_name']
            old_statistic_name = old_s.query(statistic_name). \
                filter(statistic_name.c.id == old_constant.statistic_name_id).one()
            with new.session_context() as new_s:
                new_constant = new_s.query(Constant). \
                    filter(Constant.name == old_constant.name).one_or_none()
                if new_constant:
                    copy_constant(record, old_s, old, old_constant, old_statistic_name, new_s, new_constant)
                else:
                    record.warning(f'No match for constant {old_constant.name}')
        except Exception as e:
            log_current_exception()
            record.warning(f'Error copying constant {old_constant.name}: {e}')


def copy_constant(record, old_s, old, old_constant, old_statistic_name, new_s, new_constant):
    missing = bool(new_s.query(Constant).
                   join(StatisticName).
                   outerjoin(StatisticJournal).
                   filter(Constant.id == new_constant.id,
                          StatisticJournal.id == None).count())
    if missing:
        statistic_journal = old.meta.tables['statistic_journal']
        for old_statistic_journal in old_s.query(statistic_journal). \
                filter(statistic_journal.c.source_id == old_constant.id,
                       statistic_journal.c.statistic_name_id == old_constant.statistic_name_id).all():
            copy_statistic_journal(record, old_s, old, old_statistic_name, old_statistic_journal,
                                   new_s, new_constant.statistic_name, new_constant, name=old_constant.name)
    else:
        record.warning(f'Constant {old_constant.name} already has values defined')




from logging import getLogger
from math import exp

import scipy as sp

log = getLogger(__name__)


def decay_params(period):
    # ewm has y(t) = y(t-1) * (1 - alpha) + alpha * x(t)
    # we want y(t) = y(t-1) * exp(-dt/T) + x(t)
    # clearly that won't work exactly, but we can get to within a scaling
    # so compare with y(t-1) * (1 - alpha) + alpha * x(t) * K
    # clearly K = 1/alpha
    # and exp(-dt/T) = 1 - alpha so alpha = 1 - exp(-dt/T)
    # note that this assumes evenly-sampled data of spacing dt
    decay = exp(-1 / period)
    alpha = 1 - decay
    return decay, alpha


def inplace_decay(data, column, period):
    # ewm has y(t) = y(t-1) * (1 - alpha) + alpha * x(t)
    # we want y(t) = y(t-1) * exp(-dt/T) + x(t)
    # clearly that won't work exactly, but we can get to within a scaling
    # so compare with y(t-1) * (1 - alpha) + alpha * x(t) * K
    # clearly K = 1/alpha
    # and exp(-dt/T) = 1 - alpha so alpha = 1 - exp(-dt/T)
    # note that this assumes evenly-sampled data of spacing dt
    decay = exp(-1 / period)
    alpha = 1 - decay
    log.debug(f'Exponential decay: {decay}; alpha: {alpha}; period {period} (intervals)')
    data[column] = data[column] / alpha
    data[column] = data[column].ewm(alpha=alpha, adjust=False).mean()


def fit(observed, modeled, initial_data, initial_params, evaluate, *vary,
        tol=0.01, forwards=None, backwards=None):
    '''
    Fit a model to some data.

    The model is parameterised by a namedtuple, with initial values initial_params.
    The evaluate() function takes some data and model parameters and generates the model (in the same data frame).
    The observed and modeled columns in the dataframe are then compared.

    forwards and backwards allow the model parameters to be transformed as necessary.
    forwards goes from the model to the constrained parameters; backwards reverses the transform.
    '''

    if forwards is None: forwards = lambda x: x
    if backwards is None: backwards = lambda x: x

    def objective(args):
        params = initial_params._replace(**backwards(dict(zip(vary, args))))
        data = evaluate(initial_data, params)
        return chisq(observed, modeled, data)

    result = sp.optimize.minimize(objective,
                                  [forwards(initial_params._asdict())[name] for name in vary],
                                  method='Nelder-Mead', tol=tol)
    log.debug(result)
    return initial_params._replace(**backwards(dict(zip(vary, result.x))))


def chisq(observed, modeled, data):
    delta = (data[observed] - data[modeled]).dropna()
    return sum(delta * delta)


def auto_fit(observed, modeled, initial_data, initial_params, evaluate, *vary,
             tol=0.01, forwards=None, backwards=None, max_iter=3):
    count, prev_csq = 0, None
    data = evaluate(initial_data, initial_params)
    csq = chisq(observed, modeled, data)
    while count < max_iter:
        abs_tol = tol * csq
        log.debug(f'Fitting with tol={abs_tol} (chisq {csq})')
        result = fit(observed, modeled, initial_data, initial_params, evaluate, *vary,
                     tol=abs_tol, forwards=forwards, backwards=backwards)
        data = evaluate(initial_data, result)
        csq = chisq(observed, modeled, data)
        if prev_csq is not None and abs(prev_csq - csq) / csq < tol:
            return result
        count, prev_csq = count + 1, csq


def interpolate_to_index(df, extra, *names):
    df['keep'] = True
    both = df.join(extra.loc[:, names], how='outer', sort=True)
    both.loc[both['keep'] != True, ['keep']] = False
    # todo - should be 'index' below?
    both.interpolate(method='linear', limit_area='inside', inplace=True)
    return both.loc[both['keep']].drop(columns=['keep'])




import datetime as dt
from calendar import day_name, Calendar

from bokeh.io import output_file
from bokeh.layouts import column, grid
from bokeh.models import Div, Spacer
from bokeh.plotting import show

from ch2.data import *
from ch2.lib import to_date, local_date_to_time, format_metres, format_seconds
from ch2.sql import ActivityJournal
from ch2.jupyter.decorator import template


@template
def month(month):

    f'''
    # Month: {month}
    '''

    '''
    $contents
    '''

    '''
    ## Preparation
    '''
    
    s = session('-v2')
    output_file(filename='/dev/null')
    map_size = 100
    month_start = to_date(month).replace(day=1)

    '''
    ## Generate Plot
    '''

    def days():

        for i in Calendar().iterweekdays():
            yield Div(text=f'<h2>{day_name[i]}</h2>')

        day = month_start - dt.timedelta(days=month_start.weekday())
        while day.month <= month_start.month:
            for weekday in range(7):
                if day.month == month_start.month:
                    contents = [Div(text=f'<h1>{day.strftime("%d")}</h1>')]
                    for a in s.query(ActivityJournal). \
                            filter(ActivityJournal.start >= local_date_to_time(day),
                                   ActivityJournal.finish < local_date_to_time(day + dt.timedelta(days=1))).all():
                        df = activity_statistics(s, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, activity_journal=a)
                        contents.append(map_thumbnail(map_size, map_size, df, title=False))
                        df = activity_statistics(s, ACTIVE_DISTANCE, ACTIVE_TIME, activity_journal=a)
                        contents.append(Div(
                            text=f'{format_metres(df[ACTIVE_DISTANCE][0])} {format_seconds(df[ACTIVE_TIME][0])}'))
                else:
                    contents = [Spacer()]
                yield column(contents)
                day += dt.timedelta(days=1)

    show(grid(list(days()), ncols=7))




from abc import ABC, abstractmethod
from logging import getLogger
from time import sleep

from .workers import command_root
from ..commands.args import VERBOSITY, BASE


log = getLogger(__name__)


class BaseController(ABC):

    def __init__(self, args, sys, server_cls, max_retries=5, retry_secs=3):
        self._log_level = args[VERBOSITY]
        self._base = args[BASE]
        self._sys = sys
        self.__server_cls = server_cls
        self.__max_retries = max_retries
        self.__retry_secs = retry_secs

    def status(self):
        if self._sys.exists_any_process(self.__server_cls):
            print('\n  Service running')
            self._status(True)
        else:
            print('\n  No service running')
            self._status(False)

    def _status(self, running):
        print()

    def start(self, restart=False):
        if self._sys.exists_any_process(self.__server_cls):
            log.debug('Service already running')
            if restart:
                self.stop()
            else:
                return
        log.debug('Starting remote service')
        ch2 = command_root()
        cmd, log_name = self._build_cmd_and_log(ch2)
        self._sys.run_process(self.__server_cls, cmd, log_name)
        retries = 0
        while not self._sys.exists_any_process(self.__server_cls):
            retries += 1
            if retries > self.__max_retries:
                raise Exception('Server did not start')
            sleep(self.__retry_secs)
        sleep(5)  # extra wait...
        log.info('Service started')

    @abstractmethod
    def _build_cmd_and_log(self, ch2):
        raise NotImplementedError()

    def stop(self):
        log.info('Stopping any running service')
        self._sys.delete_all_processes(self.__server_cls)
        self._cleanup()

    def _cleanup(self):
        pass

    def service(self):
        self.stop()
        log.info('Starting a local service')
        self._run()

    @abstractmethod
    def _run(self):
        raise NotImplementedError()
